import os.path
import traceback
from pathlib import Path

import pooch
import pyam
import requests
import toml


def _format_traceback_and_stdout_from_click_result(result):
    return f"{traceback.print_exception(*result.exc_info)}\n\n{result.stdout}"


def _get_infiller_download_link(filename):
    """
    Get infiller download link, intended only for use in CI
    """

    credentials_file = Path.home() / ".local" / "share" / "ixmp4" / "credentials.toml"
    with open(credentials_file, "w") as f:
        toml.dump(
            {
                "default": {
                    "username": os.environ.get("SCENARIO_EXPLORER_USER"),
                    "password": os.environ.get("SCENARIO_EXPLORER_PASSWORD"),
                }
            },
            f,
        )

    try:
        conn = pyam.iiasa.Connection()
    finally:
        # remove the yaml cred file
        credentials_file.unlink()

    infiller_url = (
        "https://db1.ene.iiasa.ac.at/ar6-public-api/rest/v2.1/files/"
        f"{filename}?redirect=false"
    )

    return requests.get(
        infiller_url,
        headers={"Authorization": f"Bearer {conn.auth.access_token}"},
    ).json()["directLink"]


def _file_available_or_downloaded(filepath, hash_exp, url):
    """
    Check if file exists (and matches expected hash) or can be downloaded

    Only intended for use in testing, but might provide inspiration for
    how to do this for others

    Parameters
    ----------
    filepath : str
        Path to file

    hash_exp : str
        Expected md5 hash

    url : str
        URL from which to download the file if it doesn't exist

    Returns
    -------
    bool
        Is the file available (or has it been downloaded hence is now
        available)?
    """
    try:
        pooch.retrieve(
            url=url,
            known_hash=f"md5:{hash_exp}",
            path=os.path.dirname(filepath),
            fname=os.path.basename(filepath),
        )
    except Exception as exc:
        # probably better ways to do this, can iterate as we use
        print(str(exc))
        return False

    return True



import logging
import os.path

import aneris
import silicone

from .checks import sanity_check_hierarchy
from .harmonization import HARMONIZATION_VARIABLES, run_harmonization
from .infilling import postprocess_infilled_for_climate, run_infilling

LOGGER = logging.getLogger(__name__)


def harmonization_and_infilling(
    df,
    key_string,
    infilling_database,
    prefix="AR6 climate diagnostics",
    instance="ar6",
    outdir="output",
    do_harmonization=True,
    # TODO: here is the downstream for potentially implementing gwp100 kyoto
):
    """
    Arguments
    ----------
    input_df : :class:`pyam.IamDataFrame`
        Input native emissions to harmonize and infill.

    inputcheck : bool
        Perform checks to remove unsuitable emissisons pathways. [Default: True]

    key_string : str
        Identifier string for writing out results. By default derived from
        the input_emissions_file string in a CLI command.

    outdir : str
        Path to output folder.

    infilling_database : str
        Path to the infiller database emissions file.

    harmonize : bool
        Perform harmonization (with False, only infilling is run). [Default: True]

    prefix : str
        Prefix used to identify the new variable names of results
        produced by the workflow.

    harmonization_instance : str
        Config string required by aneris.

    Returns
    -------
    bool
        Returns True if there are scenarios that can be run by a climate
        emulator. Those scenarios are not returned by this function, but rather
        written to the outdir. Returns False if there are no complete scenarios
        to be run.
    """
    if instance == "ar6":
        infilled_start_year = 2015
    elif instance == "sr15":
        infilled_start_year = 2010
    else:
        raise ValueError("Unknown value for instance")

    if do_harmonization:
        harmonized = run_harmonization(df, instance=instance, prefix=prefix)
    else:
        LOGGER.info("Not performing harmonization")
        harmonized = df.filter(
            year=range(infilled_start_year, 2100 + 1),
            variable=HARMONIZATION_VARIABLES,
        )
        harmonized = harmonized.rename(
            {"variable": {v: f"{prefix}|Harmonized|{v}" for v in harmonized.variable}}
        )

    if harmonized.empty:
        LOGGER.warning("No harmonized scenarios passed the checks")
        return False

    infilled, co2_infill_db, _ = run_infilling(
        harmonized,
        prefix=prefix,
        database_filepath=infilling_database,
        start_year=infilled_start_year,
    )

    # do post-processing checks after infilling
    # make sure the scenario reports until 2100
    infilled = postprocess_infilled_for_climate(
        infilled, prefix=prefix, start_year=infilled_start_year
    )

    if infilled.filter(variable="*Infilled*").empty:
        LOGGER.error("YOUR EMISSION FILE IS EMPTY AFTER INFILLING")
        return False
    else:
        # add meta string to record aneris and silicone versions
        infilled.set_meta(
            meta=(
                f"aneris (version: {aneris.__version__}), silicone (version: {silicone.__version__})"
            ),
            name="assessment-tools",
        )

        #  write out combined results of harmonization and infilling
        # write out csv file with all data (no meta)
        out_file_infilled = os.path.join(
            outdir, f"{key_string}_harmonized_infilled.csv"
        )

        LOGGER.info("Writing infilled data as csv to: %s", out_file_infilled)
        infilled.to_csv(out_file_infilled)

        # write out excel file with all data as well as meta data
        out_file_infilled_xlsx = out_file_infilled.replace(".csv", ".xlsx")
        LOGGER.info("Writing infilled data as xlsx to: %s", out_file_infilled_xlsx)
        infilled.to_excel(out_file_infilled_xlsx)

        # Sanity check for a consistent hierarchy.
        # Checks that Emissions|CO2 is the sum of AFOLU and Energy emissions
        # N.B. generally in the AR6 application total co2 co2_infill_db is empty
        if not co2_infill_db.empty:
            sanity_check_hierarchy(
                co2_infill_db,
                harmonized,
                infilled,
                out_afolu="Emissions|CO2|AFOLU",
                out_fossil="Emissions|CO2|Energy and Industrial Processes",
            )

        return True



import json
import logging

from openscm_runner.adapters import CICEROSCM

LOGGER = logging.getLogger(__name__)
DEFAULT_CICEROSCM_VERSION = "v2019vCH4"


def get_ciceroscm_configurations(
    ciceroscm_version,
    ciceroscm_probabilistic_file,
    num_cfgs,
):
    """
    Get configuration for CICERO-SCM
    """
    if CICEROSCM.get_version() != ciceroscm_version:
        # version strings for linux and windows might be different!
        raise AssertionError(CICEROSCM.get_version())

    with open(ciceroscm_probabilistic_file) as fh:
        cfgs_raw = json.load(fh)
    ciceroscm_cfgs = [c for c in cfgs_raw[:num_cfgs][:]]

    LOGGER.debug("%d total cfgs", len(ciceroscm_cfgs))

    return ciceroscm_cfgs


def ciceroscm_post_process(climate_output):
    LOGGER.info("Renaming variables")
    return climate_output



import contextlib
import logging
import os

import joblib
import numpy as np
import pandas as pd
import pyam
import scmdata
import tqdm.autonotebook as tqdman

LOGGER = logging.getLogger(__name__)


def init_logging(logger):
    # TODO: remove hard-coded level
    logger.setLevel(logging.INFO)

    # set root logger too
    logging.getLogger().setLevel(logging.INFO)
    logFormatter = logging.Formatter(
        "%(asctime)s %(name)s %(threadName)s - %(levelname)s:  %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    stdoutHandler = logging.StreamHandler()
    stdoutHandler.setFormatter(logFormatter)

    logging.getLogger().addHandler(stdoutHandler)


# drop some columns that are not required
def columns_to_basic(df):
    """
    Takes a pyam.IamDataFrame and only keeps the basic columns ["model", "scenario", "region", "variable", "unit", "year", "value"],
    discarding potential other input columns (["exclude", "meta", "subannual", "version"]).

    Parameters
    ----------
    df : :class:`pyam.IamDataFrame`
        The input data to be formatted.

    Return
    ------
    :class:`pyam.IamDataFrame`
        The input data with only the remaining basic columns.
    """
    basic_cols = ["model", "scenario", "region", "variable", "unit", "year", "value"]
    cols = df.as_pandas().columns.to_list()

    col_list = [
        ["exclude"],
        ["meta"],
        ["subannual"],
        ["version"],
        ["subannual", "version"],
        ["meta", "version"],
        ["exclude", "version"],
        ["meta", "exclude"],
        ["meta", "subannual"],
        ["meta", "version"],
        ["meta", "subannual", "version"],
        ["meta", "exclude", "version"],
        ["exclude", "meta", "subannual", "version"],
    ]

    for extra_cols in col_list:
        if sorted(cols) == sorted(basic_cols + extra_cols):
            try:
                df = pyam.IamDataFrame(df.as_pandas().drop(extra_cols, axis=1))
            except KeyError:
                raise KeyError(
                    "There are columns in the data that are unaccounted for."
                )
        else:
            # print("did not find extra columns in a common setting")
            pass

    return df


def _remove_equiv_and_hyphens_from_unit(inscmrun):
    # equiv and hyphens in units don't place nice with pint
    out = inscmrun.copy()
    out["unit"] = out["unit"].str.replace("-", "").str.replace("equiv", "")

    return out


def _add_hyphen_to_hfc4310(inscmrun):
    out = inscmrun.copy()
    out["unit"] = out["unit"].str.replace("HFC4310", "HFC43-10")

    return out


def convert_units_to_co2_equiv(df, metric):
    """
    Converts the units of gases reported in kt into Mt CO2 equivalent per year

    Uses GWP100 values from either (by default) AR5 or AR4 IPCC reports.

    Parameters
    ----------
    df : :class:`pyam.IamDataFrame`
        The input dataframe whose units need to be converted.

    metric : str
        The name of the conversion metric to use. This will usually be AR<4/5/6>GWP100.

    Return
    ------
    :class:`pyam.IamDataFrame`
        The input data with units converted.
    """
    # probably possible to do this directly with pyam too somehow, but
    # here's the scmdata version
    scmrun = scmdata.ScmRun(df.timeseries())

    # strip hyphens and equiv from inputs
    scmrun = _remove_equiv_and_hyphens_from_unit(scmrun)

    res = scmrun.convert_unit("Mt CO2/yr", context=metric).drop_meta("unit_context")

    # put back the equiv (even though it breaks pint)
    res["unit"] = "Mt CO2-equiv/yr"
    res = pyam.IamDataFrame(res.timeseries(time_axis="year"))

    return res


def convert_co2_equiv_to_kt_gas(df, var_filter, metric="AR6GWP100"):
    """
    Convert units from CO2 equivalent to kt of gas.

    Parameters
    ----------
    df : :class:`pyam.IamDataFrame`
        The input data to be converted.

    var_filter : list[str], str
        Filter to use to pick the variables to convert

    metric : str
        The name of the conversion metric to use. This will usually be AR<4/5/6>GWP100.

    Return
    ------
    :class:`pyam.IamDataFrame`
        The input data with units converted.
    """
    keep = df.filter(variable=var_filter, keep=False)
    convert = scmdata.ScmRun(df.filter(variable=var_filter))

    # strip hyphens and equiv from inputs
    convert = _remove_equiv_and_hyphens_from_unit(convert)

    converted = []
    for vdf in convert.groupby("variable"):
        variable = vdf.get_unique_meta("variable", True)
        gas = variable.split("|")[-1].replace("-", "")
        converted.append(
            vdf.convert_unit(f"kt {gas}/yr", context=metric).drop_meta("unit_context")
        )

    out = keep.append(
        _add_hyphen_to_hfc4310(scmdata.run_append(converted)).timeseries(
            time_axis="year"
        )
    )

    return out


def add_gwp100_kyoto(
    df,
    kyoto_gases=(
        "Emissions|PFC|C2F6",
        "Emissions|PFC|C6F14",
        "Emissions|PFC|CF4",
        "Emissions|CO2",
        "Emissions|CH4",
        "Emissions|HFC|HFC125",
        "Emissions|HFC|HFC134a",
        "Emissions|HFC|HFC143a",
        "Emissions|HFC|HFC227ea",
        "Emissions|HFC|HFC23",
        "Emissions|HFC|HFC32",
        "Emissions|HFC|HFC43-10",
        "Emissions|N2O",
        "Emissions|SF6",
    ),
    gwp_instance="AR6GWP100",
    prefix="",
):
    total_co2_var = f"{prefix}Emissions|CO2"

    tmp = df.copy()
    tmp.require_data(variable=total_co2_var, exclude_on_fail=True)
    calc_df = tmp.filter(exclude=False)

    # aggregate CO2 before moving on
    tmp_no_total_co2 = tmp.filter(exclude=True)
    if not tmp_no_total_co2.empty:
        LOGGER.info("Aggregating total CO2 emissions")
        tmp_no_total_co2.aggregate(total_co2_var, append=True)
        calc_df = calc_df.append(tmp_no_total_co2)

    dfp = calc_df.filter(variable=[prefix + s for s in kyoto_gases])

    if dfp.empty:
        LOGGER.warning("No Kyoto gases found with prefix %s for %s", prefix, df)
        return df

    if len(set(dfp.data["variable"])) < len(kyoto_gases):
        LOGGER.info(
            f"The input doesn't have all the variables listed in Kyoto gases. "
            f"Only the variables "
            f"{', '.join(sorted([s for s in dfp.data['variable']]))} "
            f"are included in the calculation of the GWP100."
        )

    diff = list(set(df.data["variable"]) - set(dfp.data["variable"]))
    LOGGER.info(
        f"The variables {', '.join(sorted([s for s in diff]))} "
        f"are being ignored for the calculation of the GWP100."
    )

    kyoto = (
        convert_units_to_co2_equiv(
            dfp,
            gwp_instance,
        )
        .timeseries()
        .groupby(["model", "scenario", "region", "unit"])
        .sum(min_count=1)
    )

    if gwp_instance == "AR5GWP100":
        kyoto["variable"] = prefix + "Emissions|Kyoto Gases (AR5-GWP100)"
    elif gwp_instance == "AR6GWP100":
        kyoto["variable"] = prefix + "Emissions|Kyoto Gases (AR6-GWP100)"
    else:
        raise NotImplementedError(gwp_instance)

    return pyam.IamDataFrame(pyam.concat([df, kyoto]))


def add_gwp100_kyoto_wrapper(
    df,
    prefixes=[
        "",
        "AR6 climate diagnostics|Harmonized|",
        "AR6 climate diagnostics|Infilled|",
    ],
    gwps=["AR5GWP100", "AR6GWP100"],
):
    """
    Add Kyoto GWP100 emissions

    Parameters
    ----------
    df : :class:`pyam.IamDataFrame`
        :class:`pyam.IamDataFrame` containing emissions from which the GWP sum
        should be created

    prefixes : list[str]
        List of prefixes to use for the aggregation

    gwps : list[str]
        GWPs to use for aggregation

    Returns
    -------
    :class:`pyam.IamDataFrame`
        Input emissions plus the Kyoto GWP100 equivalents
    """
    for prefix in prefixes:
        for gwp in gwps:
            LOGGER.info("Calculating %s for prefix %s", gwp, prefix)
            df = add_gwp100_kyoto(df, gwp_instance=gwp, prefix=prefix)

    return df


@contextlib.contextmanager
def parallel_progress_bar(tqdm_bar):
    """
    Context manage to patch joblib so it shows a progress bar with tqdm

    Shoutout to https://stackoverflow.com/a/58936697 for the original
    implementation of this code
    """

    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def __call__(self, *args, **kwargs):
            tqdm_bar.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_bar
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_bar.close()


def split_df(df, **filter_options):
    """
    This function splits the dataframe into model/scenario sets that fulfill
    the filter condition and those that don't.

    Parameters
    ----------
    df : :class:`pyam.IamDataFrame`
        Input scenario data

    Returns
    -------
    to_return_1 : :class:`pyam.IamDataFrame`
        Scenarios that fulfill the condition.

    to_return_2 : :class:`pyam.IamDataFrame`
        Scenarios that do not fulfill the condition.
    """
    df.set_meta("split_2", name="Split")
    mod_scens_have_lead = (
        df.filter(**filter_options)
        .data[["model", "scenario", "region"]]
        .drop_duplicates()
    )
    df.set_meta("split_1", name="Split", index=mod_scens_have_lead)
    to_return_1 = df.filter(Split="split_1")
    del to_return_1.meta["Split"]
    to_return_2 = df.filter(Split="split_1", keep=False)
    del to_return_2.meta["Split"]
    del df.meta["Split"]
    return (to_return_1, to_return_2)


def split_scenarios_into_batches(iamc_file, outdir, batch_size):
    """
    This function takes in a path to a IAMC formatted file `iamc_file` with
    scenario data, splits that into batches of size batch_size, and then saves
    those batches as CSV files in the folder `outdir`.

    Parameters
    ----------
    iamc_file : str
        Path to file that holds the scenario data that should be split.

    outdir : str
        Path to folder that will hold the produced batches.

    batch_size : int
        Up to how many scenarios should be in one batch?

    """
    init_logging(LOGGER)
    runs = pyam.IamDataFrame(iamc_file)

    model_scenario_combos = runs.meta.reset_index()[
        ["model", "scenario"]
    ].drop_duplicates()
    batches = np.array_split(
        model_scenario_combos, int(np.ceil(len(model_scenario_combos) / batch_size))
    )

    LOGGER.info(
        f"{len(model_scenario_combos)} model-scenario pairs split into {len(batches)} of size {batch_size}"
    )

    ts = runs.timeseries()

    for j, batch in tqdman.tqdm(
        enumerate(batches),
        desc="batch",
        total=len(batches),
    ):
        batch_df = pd.concat(
            [
                ts.xs(
                    (c.model, c.scenario), level=("model", "scenario"), drop_level=False
                )
                for _, c in batch.iterrows()
            ]
        )
        batch_df.to_csv(
            os.path.join(outdir, f"emissions_batch_{j + 1}.csv")
        )  # Note 1-based output files (easier for scripting)

    LOGGER.info(f"Created {len(batches)} batches")


def extract_ips(ar6_file, outdir):
    """
    This function takes an IAMC formatted file with AR6 data `iamc_file` and
    writes out only the Illustrative Pathways (IPs) in the folder `outdir`, as
    CSV.

    Parameters
    ----------
    ar6_file : str
        Path to file that holds the scenario data (e.g. full AR6 database).
        Can be either CSV or excel file.

    outdir : str
        Path to folder that will hold the output CSV file with the IPs.

    """
    IPS = (
        ("AIM/CGE 2.2", "EN_NPi2020_900f"),
        ("COFFEE 1.1", "EN_NPi2020_400f"),
        ("GCAM 5.3", "NGFS2_Current Policies"),
        ("IMAGE 3.0", "EN_INDCi2030_3000f"),
        ("MESSAGEix-GLOBIOM 1.0", "LowEnergyDemand_1.3_IPCC "),
        ("MESSAGEix-GLOBIOM_GEI 1.0", "SSP2_openres_lc_50"),
        ("REMIND-MAgPIE 2.1-4.3", "DeepElec_SSP2_ HighRE_Budg900"),
        ("REMIND-MAgPIE 2.1-4.2", "SusDev_SDP-PkBudg1000"),
        ("WITCH 5.0", "CO_Bridge"),
    )
    raw = scmdata.ScmRun(ar6_file, lowercase_cols=True)

    out = []
    for model, scenario in IPS:
        out.append(raw.filter(model=model, scenario=scenario))

    out = scmdata.run_append(out)

    out = pyam.IamDataFrame(out.timeseries(time_axis="year"))
    out.to_csv(os.path.join(outdir, "ar6_ip_data.csv"))


def _perform_operation(df, var_1, var_2, out_variable, op, raise_if_mismatch=True):
    """
    Helper function that is called in ``_diff_variables()`` and ``_add_variables()``,
    instead of silicone.multiple_infillers.infill_composite_values, because
    silicone.multiple_infillers.infill_composite_values does not check whether
    all components are included.
    """

    def _get_ts(v):
        out = df.filter(variable=v)
        if out.empty:
            return None

        return out.timeseries().reset_index("variable", drop=True)

    base = _get_ts(var_1)
    other = _get_ts(var_2)
    if base is None or other is None:
        empty = df.filter(variable="*", keep=False)
        return empty

    base, other = base.align(other)

    if op == "subtract":
        res = base - other
    elif op == "add":
        res = base + other
    else:
        raise NotImplementedError(op)

    all_nan_rows = res.isnull().all(axis=1)
    if all_nan_rows.any():
        if raise_if_mismatch:
            raise ValueError(f"Mismatched inputs: {res[all_nan_rows]}")

        res = res.loc[~all_nan_rows, :].copy()

    res["variable"] = out_variable

    return pyam.IamDataFrame(res)


def _diff_variables(df, var_1, var_2, out_variable, raise_if_mismatch=True):
    """
    Calculates the difference between two variables, and adds this new variable
    to the dataframe.

    Parameters
    ----------
    df : :class:`pyam.IamDataFrame`
        Input scenario data

    var_1 : str
        First variable in operation

    var_2 : str
        Second variable in operation

    out_variable : str
        What to call the new variable

    Returns
    -------
    :class:`pyam.IamDataFrame`
        Dataframe with added difference variable
    """
    return _perform_operation(
        df,
        var_1,
        var_2,
        out_variable,
        op="subtract",
        raise_if_mismatch=raise_if_mismatch,
    )


def _add_variables(df, var_1, var_2, out_variable, raise_if_mismatch=True):
    """
    Calculates the sum of two variables, and adds this as a new variable to the
    dataframe.

    Parameters
    ----------
    df : :class:`pyam.IamDataFrame`
        Input scenario data

    var_1 : str
        First variable in operation

    var_2 : str
        Second variable in operation

    out_variable : str
        What to call the new variable

    Returns
    -------
    :class:`pyam.IamDataFrame`
        Dataframe with added sum variable
    """
    return _perform_operation(
        df, var_1, var_2, out_variable, op="add", raise_if_mismatch=raise_if_mismatch
    )



import json
import logging

import scmdata
from openscm_runner.adapters import MAGICC7

LOGGER = logging.getLogger(__name__)
DEFAULT_MAGICC_VERSION = "v7.5.3"
DEFAULT_MAGICC_DRAWNSET = "data/magicc/0fd0f62-derived-metrics-id-f023edb-drawnset.json"


def get_magicc7_configurations(
    magicc_version,
    magicc_probabilistic_file,
    magicc_extra_config,
    num_cfgs,
    co2_and_non_co2_warming,
):
    """
    Get configuration for MAGICC7
    """
    if MAGICC7.get_version() != magicc_version:
        # version strings for linux and windows might be different!
        raise AssertionError(MAGICC7.get_version())

    with open(magicc_probabilistic_file) as fh:
        cfgs_raw = json.load(fh)

    if magicc_extra_config is not None:
        with open(magicc_extra_config) as fh:
            extra_cfgs = json.load(fh)
    else:
        extra_cfgs = {}

    common_cfgs = {
        "out_temperature": 1,
        "out_forcing": 1,
        "out_heatuptake": 1,
        "out_dynamic_vars": [
            "DAT_HEATUPTK_AGGREG",
            "DAT_CO2_CONC",
            "DAT_CH4_CONC",
            "DAT_N2O_CONC",
            "DAT_CO2_AIR2LAND_FLUX",
            "DAT_CO2_AIR2OCEAN_FLUX",
            "DAT_CO2PF_EMIS",
            "DAT_CH4PF_EMIS",
        ],
        "out_ascii_binary": "BINARY",
        "out_binary_format": 2,
        **extra_cfgs,
    }
    LOGGER.debug("Adding common_cfgs %s", common_cfgs)

    magicc7_cfgs = [
        {
            "startyear": 1750,
            "endyear": 2105,
            "run_id": c["paraset_id"],
            **c["nml_allcfgs"],
            **common_cfgs,
        }
        for c in cfgs_raw["configurations"][:num_cfgs]
    ]

    if co2_and_non_co2_warming:
        LOGGER.debug("Adding CO2 and non-CO2 warming diagnosis cfgs")
        magicc7_cfgs = [
            {**c, "rf_total_runmodus": v}
            for c in magicc7_cfgs
            for v in ["ALL", "CO2", "ANTHROPOGENIC"]
        ]
        magicc7_out_cfg = ("rf_total_runmodus",)
    else:
        magicc7_out_cfg = ()

    LOGGER.debug("%d total cfgs", len(magicc7_cfgs))

    return magicc7_cfgs, magicc7_out_cfg


def magicc7_post_process(climate_output):
    LOGGER.info("Starting MAGICC7 post-processing")

    LOGGER.info("Fixing flux variable units")
    climate_output = climate_output.convert_unit(
        "GtC/yr", variable="*Flux*CO2"
    ).convert_unit("MtCH4/yr", variable="*Flux*CH4")
    LOGGER.info("Renaming variables")
    replacements = (
        (
            "Net Land to Atmosphere Flux|CH4|Earth System Feedbacks|Permafrost",
            "Net Land to Atmosphere Flux due to Permafrost|CH4",
        ),
        (
            "Net Land to Atmosphere Flux|CO2|Earth System Feedbacks|Permafrost",
            "Net Land to Atmosphere Flux due to Permafrost|CO2",
        ),
    )
    for old, new in replacements:
        LOGGER.debug("Replacing %s with %s", old, new)
        climate_output["variable"] = climate_output["variable"].apply(
            lambda x: x.replace(old, new)
        )

    LOGGER.info("Finishing MAGICC7 post-processing")
    return climate_output


def calculate_co2_and_nonco2_warming_magicc(res):
    """
    Calculate non-CO2 warming
    """
    climate_model = res.get_unique_meta("climate_model", no_duplicates=True)
    if not climate_model.startswith("MAGICC"):
        raise AssertionError(
            "Should only have magicc results here, received "
            f"climate_model: {climate_model}"
        )

    LOGGER.debug("Removing everything except Raw Surface Temperature (GSAT)")
    res = res.filter(variable="Raw Surface Temperature (GSAT)")

    LOGGER.info("Calculating CO2 warming")
    co2_warming = res.filter(rf_total_runmodus="CO2").drop_meta("rf_total_runmodus")
    co2_warming["variable"] = co2_warming["variable"] + "|CO2"
    LOGGER.info("Calculating Non-CO2 warming")
    nonco2_warming = (
        res.filter(rf_total_runmodus="ANTHROPOGENIC")
        .subtract(
            res.filter(rf_total_runmodus="CO2"),
            op_cols={"rf_total_runmodus": "anth - co2"},
        )
        .drop_meta("rf_total_runmodus")
    )
    nonco2_warming["variable"] = nonco2_warming["variable"] + "|Non-CO2"
    LOGGER.info("Calculating Residual")
    residual_warming = (
        res.filter(rf_total_runmodus="ALL")
        .subtract(
            res.filter(rf_total_runmodus="ANTHROPOGENIC"),
            op_cols={"rf_total_runmodus": "anth - co2"},
        )
        .drop_meta("rf_total_runmodus")
    )
    residual_warming["variable"] = residual_warming["variable"] + "|Residual"

    out = scmdata.run_append(
        [co2_warming, nonco2_warming, residual_warming]
    ).convert_unit("K")

    return out



import logging
import os.path

import pandas as pd
import pandas.testing as pdt
import pyam
import scmdata
import silicone.database_crunchers
from silicone.multiple_infillers import infill_all_required_variables

from ..checks import check_negatives
from ..utils import _diff_variables, convert_co2_equiv_to_kt_gas, split_df

LOGGER = logging.getLogger(__name__)


def run_infilling(
    harmonised_df, prefix, database_filepath=None, start_year=2015, end_year=2100
):
    """
    Run infilling

    Parameters
    ----------
    harmonised_df : :class:`pyam.IamDataFrame`
        Input harmonised emissions to infill

    prefix : str
        Prefix used for the variable names

    database_filepath : str
        Path to the file which contains the infilling database

    start_year : int
        First year which should be reported in output

    end_ear : int
        Last year which should be reported in output

    Returns
    -------
    :class:`pyam.IamDataFrame`
        Infilled emissions. Output also includes the input emissions. The
        output is interpolated onto an annual timestep.
    """
    LOGGER.info("Infilling database: %s", database_filepath)
    if database_filepath is None:
        database_filepath = os.path.join(
            os.path.dirname(__file__),
            "cmip6-ssps-workflow-emissions.csv",
        )
        LOGGER.warning(
            "You are using a very simple infiller database. For any research application, it is strongly recommended to use a larger infiller database such as the AR6 infiller database instead. Please make sure to check out the documentation of this package, under 'Installation', section 'Infiller database' ."
        )

    database_filepath_cfcs = os.path.join(
        os.path.dirname(__file__), "cmip6-ssps-workflow-emissions.csv"
    )
    LOGGER.info("CFC infilling database: %s", database_filepath_cfcs)

    # We can prefix a string to the beginning of some variables to show they are infilled
    # TODO: remove hard-coding of string here
    infilled_data_prefix = f"{prefix}|Infilled"
    harmonized_prefix = f"{prefix}|Harmonized"

    # Here we specify the types of cruncher, choosing from silicone package modules.
    # This is a list of cruncher types, of the same length as the leader and required
    # variable lists
    types_of_cruncher = [
        silicone.database_crunchers.QuantileRollingWindows,
        silicone.database_crunchers.QuantileRollingWindows,
    ]

    # When infilling, what emissions data will we always have? A list of lists, the same
    # length as types_of_cruncher and qrw_variables_list
    leader_lists = [
        ["Emissions|CO2"],
        ["Emissions|CO2|Energy and Industrial Processes"],
    ]

    # What emissions data do we (sometimes) miss, or are unsure about?
    qrw_variables_list = [
        "Emissions|BC",
        "Emissions|CH4",
        "Emissions|CO2|AFOLU",
        "Emissions|CO",
        "Emissions|N2O",
        "Emissions|NH3",
        "Emissions|NOx",
        "Emissions|OC",
        "Emissions|Sulfur",
        "Emissions|VOC",
    ]

    fgas_cfc_cruncher = silicone.database_crunchers.RMSClosest
    fgas_cfc_variables_list = [
        "Emissions|HFC|HFC134a",
        "Emissions|HFC|HFC143a",
        "Emissions|HFC|HFC227ea",
        "Emissions|HFC|HFC23",
        "Emissions|HFC|HFC32",
        "Emissions|HFC|HFC43-10",
        "Emissions|HFC|HFC245ca",
        "Emissions|HFC|HFC125",
        "Emissions|SF6",
        "Emissions|PFC|CF4",
        "Emissions|PFC|C2F6",
        "Emissions|PFC|C6F14",
    ]

    cfcs_and_other_variables_cruncher = silicone.database_crunchers.RMSClosest
    cfcs_and_other_variables_list = [
        "Emissions|CCl4",
        "Emissions|CFC11",
        "Emissions|CFC113",
        "Emissions|CFC114",
        "Emissions|CFC115",
        "Emissions|CFC12",
        "Emissions|CH2Cl2",
        "Emissions|CH3Br",
        "Emissions|CH3CCl3",
        "Emissions|CH3Cl",
        "Emissions|CHCl3",
        "Emissions|HCFC141b",
        "Emissions|HCFC142b",
        "Emissions|HCFC22",
        "Emissions|HFC|HFC152a",
        "Emissions|HFC|HFC236fa",
        # 'Emissions|HFC|HFC245fa',
        "Emissions|HFC|HFC365mfc",
        "Emissions|Halon1202",
        "Emissions|Halon1211",
        "Emissions|Halon1301",
        "Emissions|Halon2402",
        "Emissions|NF3",
        "Emissions|PFC|C3F8",
        "Emissions|PFC|C4F10",
        "Emissions|PFC|C5F12",
        "Emissions|PFC|C7F16",
        "Emissions|PFC|C8F18",
        "Emissions|PFC|cC4F8",
        "Emissions|SO2F2",
    ]
    # Timesteps to report data on
    output_timesteps = list(range(start_year, end_year + 1, 1))

    # Convert the hfc/pfc units from CO2 equiv to mass of gas
    convert_hfc_units = True
    convert_pfc_units = True

    # Check that options make sense
    if len(leader_lists) != len(types_of_cruncher):
        raise AssertionError(
            "Number of lead variables doesn't match number of crunchers"
        )

    LOGGER.info("Loading infilling database")
    database = scmdata.ScmRun(database_filepath, lowercase_cols=True)

    LOGGER.info("Loading infilling database cfcs")
    database_cfcs = scmdata.ScmRun(database_filepath_cfcs, lowercase_cols=True)

    if len(database["region"].unique()) > 1:
        raise AssertionError(
            "Different regions should be "
            f"infilled separately. Your database has regions {database.regions()}"
        )

    to_fill_orig = harmonised_df.copy()
    if len(to_fill_orig.region) > 1:
        raise AssertionError(
            f"Different regions should be infilled separately. You are infilling {to_fill_orig.regions()}"
        )

    if to_fill_orig["region"][0] != database["region"][0]:
        raise AssertionError(
            "The cruncher data and the infilled data have different regions."
        )

    # Perform situation-specific data cleansing on variable names
    # ___________________________________________________________
    LOGGER.info("Preparing infilling databases")
    database["variable"] = database["variable"].apply(
        lambda x: f"Emissions{x.split('Emissions')[-1]}"
    )
    if not database["variable"].str.startswith("Emissions").all():
        raise AssertionError("Something fishy going on with prefix handling")

    for db, name in ((database, "database"), (database_cfcs, "database_cfcs")):
        late_start = min(db["year"]) > min(output_timesteps)
        early_finish = max(db["year"]) < max(output_timesteps)
        if late_start or early_finish:
            raise AssertionError(
                f"Database {name} does not extend far enough to be used for infilling"
            )

    database = pyam.IamDataFrame(database.timeseries(time_axis="year")).interpolate(
        output_timesteps, inplace=False
    )

    database_cfcs = pyam.IamDataFrame(
        database_cfcs.filter(year=output_timesteps).timeseries(time_axis="year")
    ).interpolate(output_timesteps, inplace=False)

    # Handle CO2 reporting
    # ____________________
    co2_total = "Emissions|CO2"
    co2_energy = "Emissions|CO2|Energy and Industrial Processes"
    co2_afolu = "Emissions|CO2|AFOLU"

    # if anything has total CO2 and energy, calculate AFOLU CO2 as difference
    # if anything has total CO2 and AFOLU, calculate energy CO2 as difference
    for has, missing in ((co2_energy, co2_afolu), (co2_afolu, co2_energy)):
        _, missing_sector = split_df(
            harmonised_df, variable=f"{harmonized_prefix}|{missing}"
        )

        if not missing_sector.empty:
            sector_inferred = _diff_variables(
                missing_sector,
                f"{harmonized_prefix}|{co2_total}",
                f"{harmonized_prefix}|{has}",
                f"{harmonized_prefix}|{missing}",
                raise_if_mismatch=False,
            )

            if not sector_inferred.empty:
                harmonised_df = harmonised_df.append(sector_inferred, inplace=False)

    # The remainder of the program should be fairly standardised
    # __________________________________________________________
    still_to_infill = harmonised_df
    infilled = None
    for ind, lead in enumerate(leader_lists):
        LOGGER.info("Infilling using %s as the lead variable", lead)
        to_infill, still_to_infill = split_df(
            still_to_infill, variable=harmonized_prefix + "|" + lead[0]
        )

        if not to_infill:
            LOGGER.info("Nothing to infill")
            continue

        LOGGER.info("Interpolating data to infill")
        to_infill = to_infill.interpolate(output_timesteps, inplace=False)

        if lead == ["Emissions|CO2"]:
            _, missing_energy = split_df(
                to_infill, variable=f"{harmonized_prefix}|{co2_energy}"
            )
            if not missing_energy.empty:
                # infill CO2 energy for scenarios that have CO2 total but no energy

                if co2_afolu in missing_energy.variable:
                    # We have somehow ended up with CO2 total and AFOLU but
                    # not energy, even though we should have calculated energy
                    # from total and AFOLU above
                    raise AssertionError(
                        "CO2 energy should have been calculated earlier"
                    )

                infilled_energy = _infill_variables(
                    types_of_cruncher[ind],
                    [co2_energy],
                    missing_energy,
                    database,
                    lead,
                    output_timesteps,
                    old_prefix=harmonized_prefix,
                )

                # calculate CO2 AFOLU as difference between energy and total
                # to preserve total and harmonisation
                afolu_inferred = _diff_variables(
                    infilled_energy,
                    co2_total,
                    co2_energy,
                    co2_afolu,
                )

                infilled = _add_to_infilled(infilled, infilled_energy)
                infilled = _add_to_infilled(infilled, afolu_inferred)

        for cruncherh, variables, database_here in (
            (types_of_cruncher[ind], qrw_variables_list, database),
            (fgas_cfc_cruncher, fgas_cfc_variables_list, database),
            (
                cfcs_and_other_variables_cruncher,
                cfcs_and_other_variables_list,
                database_cfcs,
            ),
        ):
            infilled_variables = _infill_variables(
                cruncherh,
                variables,
                to_infill,
                database_here,
                lead,
                output_timesteps,
                old_prefix=harmonized_prefix,
            )
            infilled = _add_to_infilled(infilled, infilled_variables)

    if infilled is None:
        raise ValueError("No infilling occured. Check input emissions")

    LOGGER.info("Converting HFC/PFC units back from CO2-equivalent")
    if convert_hfc_units:
        infilled = convert_co2_equiv_to_kt_gas(infilled, "*|HFC|*")
    if convert_pfc_units:
        infilled = convert_co2_equiv_to_kt_gas(infilled, "*|PFC|*")

    # We want to ensure that the original data is preserved as it was entered and that a
    # complete copy of the completed data is tagged as such.
    if infilled_data_prefix:
        infilled = infilled.data
        infilled["variable"] = infilled_data_prefix + "|" + infilled["variable"]
        infilled = pyam.IamDataFrame(infilled)

    # We also want a clone of the original preserved data, unless nothing changed
    if not pyam.compare(infilled, to_fill_orig).empty:
        infilled = infilled.append(to_fill_orig)
    else:
        LOGGER.warning(
            "The data was already complete and in the correct format. It has not changed"
        )

    out = infilled.filter(
        variable="*Kyoto*", keep=False
    )  # ad-hoc fix to ensure no Kytoto GWP100 coming in the infilling process

    # remove any infilled CO2 total but return so we can use it in the tests (
    # this is a hack which should be removed in future but is important now
    # while our functions are still pretty big)
    co2_total = out.filter(variable="*Infilled*Emissions|CO2")
    # Get "Emissions|CO2" of the infiller database of the used models, scenarios and regions
    co2_infiller_db = database.filter(
        variable="Emissions|CO2",
        model=list(set(out["model"])),
        scenario=list(set(out["scenario"])),
        region=list(set(out["region"])),
    )
    out = out.filter(variable="*Infilled*Emissions|CO2", keep=False)

    return out, co2_infiller_db, co2_total


def _infill_variables(
    cruncher, variables, to_infill, db, lead, output_timesteps, old_prefix
):
    """
    Start core infilling run, using silicone.multiple_infillers.infill_all_required_variables

    Parameters
    ----------
    cruncher : str
        Infilling method

    variables : list
        List of required variables to infill

    to_infill : :class:`pyam.IamDataFrame`
        Emissions data to be infilled

    db : :class:`pyam.IamDataFrame`
        The infilling database

    lead : str
        Lead infiller variable

    output_timesteps : list
        List of required output timesteps

    old_prefix : str
        The prefix of harmonized emissions, which will be replaced by a prefix
        for the infilled emissions variables.

    Returns
    -------
    :class:`pyam.IamDataFrame`
        Infilled emissions.
    """
    LOGGER.info("Infilling using cruncher %s", cruncher)
    LOGGER.info("Infilling %s", variables)
    infilled_variables = infill_all_required_variables(
        to_infill.copy(),
        db,
        lead,
        required_variables_list=variables,
        cruncher=cruncher,
        output_timesteps=output_timesteps,
        to_fill_old_prefix=old_prefix,
        check_data_returned=True,
    )
    del infilled_variables.meta["already_filled"]

    return infilled_variables


def _add_to_infilled(infilled, infilled_variables):
    """
    Helper function which takes two dataframes, and binds them together.

    Parameters
    ----------
    infilled : :class:`pyam.IamDataFrame`
        An IamDataFrame.
    infilled_variables : str, None
        An IamDataFrame.

    Returns
    -------
    :class:`pyam.IamDataFrame`
        An IamDataFrame.
    """
    if infilled is not None:
        infilled = infilled.timeseries()
        infilled_variables = infilled_variables.timeseries()
        keep_idx = infilled_variables.index.difference(infilled.index)
        if keep_idx.empty:
            LOGGER.debug("No timeseries infilled")
            return infilled

        infilled = pyam.IamDataFrame(
            pd.concat([infilled, infilled_variables.loc[keep_idx]])
        )

    else:
        infilled = infilled_variables

    return infilled


def load_csv_or_xlsx_for_one_region(file, sheet):
    """
    This loads data from either a csv or an xls(x) file and checks that it has only a
    single region.

    Parameters
    ----------
    file : str, None
        The (relative or absolute) file address of the file to load
    sheet : str, None
        If file is a csv, use None. Otherwise, the sheet name to read from the file.

    Returns
    -------
    :class:`pyam.IamDataFrame`
        The information in the file.
    """
    # TODO: loading databases with pyam is slow because of reshaping
    if sheet:
        my_df = pyam.IamDataFrame(file, sheet_name=sheet).data
    else:
        my_df = pyam.IamDataFrame(file).data

    if my_df["region"].nunique() > 1:
        raise AssertionError("Multiple regions must be infilled separately")

    return my_df


def postprocess_infilled_for_climate(df_infilled, prefix, start_year=2015):
    """
    Helper function that takes a set of infilled emissions data, adds a provided
    prefix, and filters out the scenarios that are appropriately infilled by checking
    both for a minimum set of years of data and checking that there are no non-co2
    negative emissions caused by the infilling method.

    Parameters
    ----------
    df_infilled : :class:`pyam.IamDataFrame`
        Infilled data.
    prefix : str
        Prefix used for the variable names

    Returns
    -------
    :class:`pyam.IamDataFrame`
        Checked and reformatted infilled data.
    """
    LOGGER.info("Post-processing for climate models")

    def _infilled_name(v):
        return "|".join([prefix, "Infilled", v])

    required_vars = {
        _infilled_name("Emissions|BC"),
        _infilled_name("Emissions|CH4"),
        # _infilled_name("Emissions|CO2"),
        # _infilled_name("Emissions|CO2|Other"),
        # _infilled_name("Emissions|CO2|Waste"),
        _infilled_name("Emissions|CH4"),
        _infilled_name("Emissions|CO"),
        _infilled_name("Emissions|CO2|AFOLU"),
        _infilled_name("Emissions|CO2|Energy and Industrial Processes"),
        # _infilled_name("Emissions|F-Gases"),
        # _infilled_name("Emissions|HFC"),
        _infilled_name("Emissions|HFC|HFC125"),
        _infilled_name("Emissions|HFC|HFC134a"),
        _infilled_name("Emissions|HFC|HFC143a"),
        _infilled_name("Emissions|HFC|HFC227ea"),
        _infilled_name("Emissions|HFC|HFC23"),
        _infilled_name("Emissions|HFC|HFC245ca"),  # not in historical dataset (RCMIP)
        # _infilled_name("Emissions|HFC|HFC245fa"),
        _infilled_name("Emissions|HFC|HFC32"),
        _infilled_name("Emissions|HFC|HFC43-10"),
        _infilled_name("Emissions|N2O"),
        _infilled_name("Emissions|NH3"),
        _infilled_name("Emissions|NOx"),
        _infilled_name("Emissions|OC"),
        # _infilled_name("Emissions|PFC"),
        _infilled_name("Emissions|PFC|C2F6"),
        _infilled_name("Emissions|PFC|C6F14"),
        _infilled_name("Emissions|PFC|CF4"),
        _infilled_name("Emissions|SF6"),
        _infilled_name("Emissions|Sulfur"),
        _infilled_name("Emissions|VOC"),
    }
    # mark the scenarios that are not sufficiently infilled for climate assessment, and filter scenarios
    required_years = list(range(start_year, 2100 + 1))
    LOGGER.info("Checking infilled results have required years and variables")
    for (model, scenario), msdf in (
        df_infilled.filter(year=required_years)
        .timeseries()
        .groupby(["model", "scenario"])
    ):
        has_all_years = set(msdf.columns) == set(required_years)
        has_all_variables = not required_vars - set(
            msdf.index.get_level_values("variable")
        )
        if not (has_all_years and has_all_variables):
            LOGGER.info("Removing %s %s", model, scenario)
            df_infilled = df_infilled.filter(model=model, scenario=scenario, keep=False)

    # in case non-CO2 negatives have been introduced here, we want to know!
    LOGGER.info("Check that there are no non-CO2 negatives introduced by infilling")
    df_infilled_negatives_checked = check_negatives(df_infilled, prefix=f"{prefix}*")
    pdt.assert_frame_equal(
        df_infilled.timeseries().T,
        df_infilled_negatives_checked.timeseries().T,
        check_like=True,
    )

    return df_infilled



import logging
import os

import aneris.convenience
import pandas as pd
import pyam
import scmdata
import tqdm
from joblib import Parallel, delayed

from climate_assessment.checks import remove_rows_with_zero_in_harmonization_year

from ..utils import parallel_progress_bar

LOGGER = logging.getLogger(__name__)


here = os.path.dirname(os.path.realpath(__file__))


HARMONIZATION_VARIABLES = [
    "Emissions|BC",
    "Emissions|PFC|C2F6",
    "Emissions|PFC|C6F14",
    "Emissions|PFC|CF4",
    "Emissions|CO",
    "Emissions|CO2",
    "Emissions|CO2|AFOLU",
    "Emissions|CO2|Energy and Industrial Processes",
    "Emissions|CH4",
    "Emissions|F-Gases",
    "Emissions|HFC",
    "Emissions|HFC|HFC125",
    "Emissions|HFC|HFC134a",
    "Emissions|HFC|HFC143a",
    "Emissions|HFC|HFC227ea",
    "Emissions|HFC|HFC23",
    # 'Emissions|HFC|HFC245ca',  # all nan in historical dataset (RCMIP)
    # "Emissions|HFC|HFC245fa",  # not in historical dataset (RCMIP)
    "Emissions|HFC|HFC32",
    "Emissions|HFC|HFC43-10",
    "Emissions|N2O",
    "Emissions|NH3",
    "Emissions|NOx",
    "Emissions|OC",
    "Emissions|PFC",
    "Emissions|SF6",
    "Emissions|Sulfur",
    "Emissions|VOC",
]


# get path for data input folder
def getpath(f, data_folder=here):
    return os.path.join(here, f)


def postprocessing(harmonized_results, prefix):
    """
    Helper function that makes sure that the harmonization prefix is added.
    """
    # hotfix prefix|suffix for harmonized emissions output
    # strip suffix
    harmonized_results = harmonized_results.data
    harmonized_results["variable"] = harmonized_results["variable"].map(
        lambda x: x.replace("|Harmonized", "")
    )
    #    replace prefix
    harmonized_results["variable"] = (
        harmonized_results["variable"]
        .map(lambda x: x.replace(prefix, f"{prefix}|Harmonized"))
        .astype(str)
    )
    harmonized_results = pyam.IamDataFrame(harmonized_results)

    return harmonized_results


def add_year_historical_percentage_offset(df, dfhist, yr=2015, low_yr=2010):
    """
    In the case that the chosen harmonization year (yr, default=2015) is not reported,
    this function uses an earlier year (low_yr, default=2010) to derive the
    historical offset for harmonization.
    It takes the relative difference in `low_yr`, and assumes that this relative
    difference stays the same until `yr`.
    """
    if yr not in df.columns:
        df[yr] = None
        df[yr] = pd.to_numeric(df[yr])

    df2015 = df[~df[yr].isnull()]
    dfno2015 = df[df[yr].isnull()].copy()

    if low_yr in dfno2015.columns:
        dfhist_low = dfhist[low_yr].reset_index(["model", "scenario"], drop=True)
        dfhist_yr = dfhist[yr].reset_index(["model", "scenario"], drop=True)

        dfno2015_low = dfno2015[[low_yr]]

        relative_diff = (
            dfno2015_low.subtract(dfhist_low, axis=0)
            .divide(dfhist_low, axis=0)
            .dropna()
        )
        if relative_diff.shape[0] != dfno2015.shape[0]:
            raise AssertionError("Some data will not get adjusted properly")

        fill_values = (
            relative_diff.multiply(dfhist_yr, axis=0)
            .add(dfhist_yr, axis=0)
            .dropna()
            .rename({low_yr: yr}, axis="columns")
        )
        if fill_values.shape[0] != dfno2015.shape[0]:
            raise AssertionError("Some data will not get adjusted properly")

        dfno2015[yr] = fill_values[yr].reorder_levels(dfno2015.index.names)
        df = pd.concat([df2015, dfno2015])
    else:
        if not dfno2015.empty:
            raise KeyError(f"{low_yr} not in `dfno2015`")

    return df


def run_harmonization(df, instance, prefix):
    """
    Run harmonization.
    Hamronization method overrides by specific species are set within this function too.

    Parameters
    ----------
    df : :class:`pyam.IamDataFrame`
        Input native emisisons in IAMC format.
    instance : str
        String used to choose what historical data to use.
    prefix : str
        Prefix used for the variable names

    Returns
    -------
    :class:`pyam.IamDataFrame`
        Harmonized scenarios.
    """
    LOGGER.info(f"Using {instance} instance for harmonization")

    df_hist = scmdata.ScmRun(
        getpath("history_" + instance + ".csv"), lowercase_cols=True
    )

    LOGGER.debug("Emissions to harmonize %s", HARMONIZATION_VARIABLES)

    not_harmonize = set(df.variable) - set(HARMONIZATION_VARIABLES)
    LOGGER.info("Not harmonizing %s", not_harmonize)

    df = df.filter(variable=HARMONIZATION_VARIABLES)
    if df.empty:
        return df

    # TODO: remove this hard-coding
    if instance.startswith("sr15"):
        harmonization_year = 2010
    else:
        harmonization_year = 2015

    LOGGER.info("harmonization_year %s", harmonization_year)

    scmrun = scmdata.ScmRun(df)
    LOGGER.info("Stripping equivalent units for processing")
    scmrun["unit"] = scmrun["unit"].str.replace("-equiv", "").str.replace("-", "")

    scenarios = scmrun.copy()

    LOGGER.info("Creating pd.DataFrame's for aneris")
    scenarios = scenarios.timeseries(time_axis="year")

    df_hist["variable"] = df_hist["variable"].apply(
        lambda x: x.replace(f"{prefix}|", "").replace("|Unharmonized", "")
    )
    df_hist["unit"] = df_hist["unit"].str.replace("-equiv", "").str.replace("-", "")
    history = df_hist.filter(year=range(1990, 2020)).timeseries(time_axis="year")

    # TODO: remove hard-coding
    historical_offset_add_year = 2015
    historical_offset_base_year = 2010
    if harmonization_year == historical_offset_add_year:
        LOGGER.info(
            "Adding %s values based on historical percentage offset from %s",
            historical_offset_add_year,
            historical_offset_base_year,
        )
        scenarios = add_year_historical_percentage_offset(
            scenarios,
            history,
            yr=historical_offset_add_year,
            low_yr=historical_offset_base_year,
        )

    LOGGER.info("Interpolating onto output timesteps")
    # TODO: remove hard-coded end year
    output_timesteps = range(harmonization_year, 2100 + 1)
    LOGGER.debug("output_timesteps %s", output_timesteps)
    scenarios = pyam.IamDataFrame(scenarios).interpolate(output_timesteps)
    scenarios = remove_rows_with_zero_in_harmonization_year(
        scenarios,
        filename="dropped_rows",
        harmonization_year=harmonization_year,
    )  # note: this process is repeated after harmonization. Before is slightly nicer, but not enough.

    # TODO: remove hard-coding
    overrides = pd.DataFrame(
        [
            #     {'method': 'default_aneris_tree', 'variable': 'Emissions|BC'}, # depending on the decision tree in aneris/method.py
            {
                "method": "reduce_ratio_2150_cov",
                "variable": "Emissions|PFC",
            },  # high historical variance (cov=16.2)
            {
                "method": "reduce_ratio_2150_cov",
                "variable": "Emissions|PFC|C2F6",
            },  # high historical variance (cov=16.2)
            {
                "method": "reduce_ratio_2150_cov",
                "variable": "Emissions|PFC|C6F14",
            },  # high historical variance (cov=15.4)
            {
                "method": "reduce_ratio_2150_cov",
                "variable": "Emissions|PFC|CF4",
            },  # high historical variance (cov=11.2)
            {
                "method": "reduce_ratio_2150_cov",
                "variable": "Emissions|CO",
            },  # high historical variance (cov=15.4)
            {
                "method": "reduce_ratio_2080",
                "variable": "Emissions|CO2",
            },  # always ratio method by choice
            {
                "method": "reduce_offset_2150_cov",
                "variable": "Emissions|CO2|AFOLU",
            },  # high historical variance, but using offset method to prevent diff from increasing when going negative rapidly (cov=23.2)
            {
                "method": "reduce_ratio_2080",  # always ratio method by choice
                "variable": "Emissions|CO2|Energy and Industrial Processes",
            },
            #     {'method': 'default_aneris_tree', 'variable': 'Emissions|CH4'}, # depending on the decision tree in aneris/method.py
            {
                "method": "constant_ratio",
                "variable": "Emissions|F-Gases",
            },  # basket not used in infilling (sum of f-gases with low model reporting confidence)
            {
                "method": "constant_ratio",
                "variable": "Emissions|HFC",
            },  # basket not used in infilling (sum of subset of f-gases with low model reporting confidence)
            {
                "method": "constant_ratio",
                "variable": "Emissions|HFC|HFC125",
            },  # minor f-gas with low model reporting confidence
            {
                "method": "constant_ratio",
                "variable": "Emissions|HFC|HFC134a",
            },  # minor f-gas with low model reporting confidence
            {
                "method": "constant_ratio",
                "variable": "Emissions|HFC|HFC143a",
            },  # minor f-gas with low model reporting confidence
            {
                "method": "constant_ratio",
                "variable": "Emissions|HFC|HFC227ea",
            },  # minor f-gas with low model reporting confidence
            {
                "method": "constant_ratio",
                "variable": "Emissions|HFC|HFC23",
            },  # minor f-gas with low model reporting confidence
            {
                "method": "constant_ratio",
                "variable": "Emissions|HFC|HFC32",
            },  # minor f-gas with low model reporting confidence
            {
                "method": "constant_ratio",
                "variable": "Emissions|HFC|HFC43-10",
            },  # minor f-gas with low model reporting confidence
            #     {'method': 'default_aneris_tree', 'variable': 'Emissions|N2O'}, # depending on the decision tree in aneris/method.py
            #     {'method': 'default_aneris_tree', 'variable': 'Emissions|NH3'}, # depending on the decision tree in aneris/method.py
            #     {'method': 'default_aneris_tree', 'variable': 'Emissions|NOx'}, # depending on the decision tree in aneris/method.py
            {
                "method": "reduce_ratio_2150_cov",
                "variable": "Emissions|OC",
            },  # high historical variance (cov=18.5)
            {
                "method": "constant_ratio",
                "variable": "Emissions|SF6",
            },  # minor f-gas with low model reporting confidence
            #     {'method': 'default_aneris_tree', 'variable': 'Emissions|Sulfur'}, # depending on the decision tree in aneris/method.py
            {
                "method": "reduce_ratio_2150_cov",
                "variable": "Emissions|VOC",
            },  # high historical variance (cov=12.0)
        ]
    )
    LOGGER.info("Harmonisation overrides:\n%s", overrides)

    scenarios = scenarios.filter(year=output_timesteps).timeseries()
    with parallel_progress_bar(tqdm.tqdm(desc="Harmonisation")):
        LOGGER.info("Harmonising in parallel")
        # TODO: remove hard-coding of n_jobs
        scenarios_harmonized = Parallel(n_jobs=-1)(
            delayed(aneris.convenience.harmonise_all)(
                msdf,
                history=history,
                harmonisation_year=harmonization_year,
                overrides=overrides,
            )
            for _, msdf in scenarios.groupby(["model", "scenario"])
        )

    LOGGER.info("Hacking around some regression in aneris - pyam stack")

    def drop_broken_stuff(indf):
        out = indf.copy()
        idx_length = len(out.index.names)
        drop_levels = list(range(idx_length // 2, idx_length))
        out.index = out.index.droplevel(drop_levels)

        return out

    scenarios_harmonized = [drop_broken_stuff(s) for s in scenarios_harmonized]

    LOGGER.info("Combining results")
    scenarios_harmonized = pd.concat(scenarios_harmonized).reset_index()

    LOGGER.info("Putting equiv back in units")
    equiv_rows = scenarios_harmonized["variable"].isin(["Emissions|HFC"])
    scenarios_harmonized.loc[equiv_rows, "unit"] = scenarios_harmonized.loc[
        equiv_rows, "unit"
    ].str.replace("/yr", "-equiv/yr")

    scenarios_harmonized["variable"] = (
        f"{prefix}|Harmonized|" + scenarios_harmonized["variable"]
    )

    LOGGER.info("Converting to IamDataFrame")
    scenarios_harmonized = pyam.IamDataFrame(scenarios_harmonized)

    # TODO: perhaps move to checks, if this is possible before doing historical offset and then interpolation
    # NB: doing this here in the workflow will fail if there are scenarios being harmonized to historical zero emissions!
    LOGGER.info("CHECK: if some emissions are zero in the harmonization year.")
    scenarios_harmonized = remove_rows_with_zero_in_harmonization_year(
        scenarios_harmonized,
        filename="dropped_rows",
        harmonization_year=harmonization_year,
    )
    if ((scenarios_harmonized.filter(year=2015).as_pandas())["value"] == 0).any():
        raise AssertionError("Some scenarios have zero in the harmonization year!")

    # TODO: some logic to check that CO2 total was handled properly?

    return scenarios_harmonized



import logging
import os.path
from concurrent.futures import ProcessPoolExecutor, as_completed

import click
import pandas as pd
import pyam
from tqdm import tqdm

from .checks import (
    infiller_vetting,
    perform_input_checks,
    sanity_check_bounds_kyoto_emissions,
    sanity_check_comparison_kyoto_gases,
)
from .climate import DEFAULT_MAGICC_VERSION, climate_assessment
from .climate.post_process import check_hist_warming_period
from .harmonization import run_harmonization
from .harmonization_and_infilling import harmonization_and_infilling
from .infilling import postprocess_infilled_for_climate, run_infilling
from .postprocess import do_postprocess
from .utils import (
    _add_variables,
    add_gwp100_kyoto_wrapper,
    columns_to_basic,
    extract_ips,
    init_logging,
    split_df,
    split_scenarios_into_batches,
)

LOGGER = logging.getLogger(__name__)

input_emissions_file_arg = click.argument(
    "input_emissions_file",
    required=True,
    type=click.Path(exists=True, dir_okay=False, readable=True, resolve_path=True),
)
outdir_arg = click.argument(
    "outdir",
    required=True,
    type=click.Path(exists=True, file_okay=False, writable=True, resolve_path=True),
)
output_files_arg = click.argument(
    "rawoutput_files",
    nargs=-1,
    type=click.Path(exists=True, file_okay=True, resolve_path=True),
)
harmonizedinfilledemissions_arg = click.argument(
    "harmonizedinfilledemissions",
    type=click.Path(exists=True, dir_okay=False, readable=True, resolve_path=True),
)
n_workers_option = click.option(
    "--n_workers",
    help="Number of workers",
    default=8,
    type=int,
)
output_file_option = click.option(
    "--output-file",
    help="Filename where the combined metadata is output",
    required=True,
)
input_check_option = click.option(
    "--inputcheck/--no-inputcheck",
    help="Check input before running the workflow",
    required=False,
    default=True,
    show_default=True,
)
infilling_database_option = click.option(
    "--infilling-database",
    help="File to use as the infilling database",
    required=False,  # defaults to infiller database used for ar6
    default=os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            "infilling",
            "cmip6-ssps-workflow-emissions.csv",
        )
    ),
    type=click.Path(exists=True, file_okay=True, readable=True, resolve_path=True),
    show_default=True,
)
harmonization_instance_option = click.option(
    "--harmonization-instance",
    help="Harmonisation settings to use",
    required=False,  # defaults to rcmip historical data as used for ar6
    default="ar6",
    type=click.Choice(["ar6", "sr15"], case_sensitive=True),
    show_default=True,
)
model_option = click.option(
    "--model",
    help="Climate model to run scenarios with",
    required=False,
    default="magicc",
    type=str,
    show_default=True,
)
model_version_option = click.option(
    "--model-version",
    help="Expected model version (just used to check env is working as intended)",
    default=DEFAULT_MAGICC_VERSION,
    type=str,
    show_default=True,
)
probabilistic_file_option = click.option(
    "--probabilistic-file",
    help="json file containing climate model probabilistic config",
    required=True,  # currenlty no default here, users need to locate it
    type=click.Path(exists=True, dir_okay=False, readable=True, resolve_path=True),
)
magicc_extra_config_option = click.option(
    "--magicc-extra-config",
    help="File containing additional MAGICC config",
    required=False,
    type=click.Path(exists=True, dir_okay=False, readable=True, resolve_path=True),
)
fair_extra_config_option = click.option(
    "--fair-extra-config",
    help="File containing additional FaIR config",
    required=False,  # required when running FaIR
    type=click.Path(exists=True, dir_okay=False, readable=True, resolve_path=True),
)
hist_warming_option = click.option(
    "--historical-warming",
    help="Historical warming estimate in kelvin",
    required=False,  # defaults to WGI estimated historical warming
    default=0.85,
    type=float,
    show_default=True,
)
hist_warming_ref_period_option = click.option(
    "--historical-warming-reference-period",
    help="Reference period used when calculating historical warming",
    required=False,
    default="1850-1900",
    type=str,
    show_default=True,
)
hist_warming_eval_period_option = click.option(
    "--historical-warming-evaluation-period",
    help="Evaluation period used when calculating historical warming",
    required=False,
    default="1995-2014",
    type=str,
    show_default=True,
)
num_cfgs_option = click.option(
    "--num-cfgs",
    help="Number of climate model configs to run",
    default=600,
    type=int,
    show_default=True,
)
test_run_option = click.option(
    "--test-run/--no-test-run",
    help="Make this run a test run (doesn't check that historical warming"
    "from the climate models comes out consistently)",
    default=False,
    type=bool,
    show_default=True,
)
scenario_batch_size_option = click.option(
    "--scenario-batch-size",
    help="Number of scenarios to run at a time",
    required=False,
    default=20,
    type=int,
    show_default=True,
)
save_raw_climate_output_option = click.option(
    "--save-raw-climate-output/--dont-save-raw-climate-output",
    help="Save raw climate output to disk",
    required=False,
    default=False,
    type=bool,
    show_default=True,
)
categorisation_option = click.option(
    "--categorisation/--no-categorisation",
    help="Add temperature category to meta data",
    required=False,
    default=True,
    show_default=True,
)
report_completeness_option = click.option(
    "--reporting-completeness-categorisation/"
    "--no-reporting-completeness-categorisation",
    help="Add an emissions reporting completeness category to meta data",
    required=False,
    default=False,
    type=bool,
    show_default=True,
)
nonco2_warming_option = click.option(
    "--co2-and-non-co2-warming/--no-co2-and-non-co2-warming",
    help="Include CO2 and non-CO2 warming assessment",
    required=False,
    default=False,
    type=bool,
    show_default=True,
)
prefix_all_out_variables_option = click.option(
    "--prefix", help="Prefix for all output variables", required=True
)
prefix_option = click.option(
    "--prefix",
    help="Prefix for all variable names",
    required=False,
    default="AR6 climate diagnostics",
    type=str,
    show_default=True,
)
gwp_option = click.option(
    "--gwp/--no-gwp",
    help="Add Kyoto Gases in GWP100 to output",
    required=False,
    default=True,
    type=bool,
    show_default=True,
)
gwp_def_false_option = click.option(
    "--gwp/--no-gwp",
    help="Add Kyoto Gases in GWP100 to output",
    default=False,
    type=bool,
    show_default=True,
)
kyoto_ghgs_option = click.option(
    "--kyoto-ghgs", help="Add Kyoto GHGs to output", default=True
)
postprocess_option = click.option(
    "--postprocess/--no-postprocess",
    help="Run postprocessing steps",
    default=True,
    type=bool,
    show_default=True,
)
harmonize_option = click.option(
    "--harmonize/--dont-harmonize",
    help="Do harmonization",
    default=True,
    type=bool,
    show_default=True,
)
save_csv_combined_output_option = click.option(
    "--save-csv-combined-output",
    help="Write CSV output with combined climate output and emissions",
    is_flag=True,
    required=False,
    default=False,
    type=bool,
    show_default=True,
)


def _setup_logging(logger):
    """
    Set up logging preferences. This removes unnecessary logger warnings from
    packages that the workflow depends on, most notably pyam. It also sets the
    general logger level to INFO.

    Parameters
    ----------
    logger : :class:`logging.getLogger`
        Input Logger instance created by logging.getLogger(name).
    """
    init_logging(logger)

    logger.info("Silencing pyam loggers")
    logging.getLogger("pyam.core").setLevel(logging.CRITICAL)
    logging.getLogger("pyam.utils").setLevel(logging.CRITICAL)


def _get_key_string_and_log_outdir(input_emissions_file, outdir, logger):
    # TODO: remove need to parse in logger once we've cleaned things up
    logger.info("Outputs will be saved in: %s", outdir)
    key_string = os.path.splitext(os.path.basename(input_emissions_file))[0]
    logger.info("Outputs will be saved with the ID: %s", key_string)

    return key_string


def _load_emissions_convert_to_basic(input_emissions_file, logger):
    # TODO: remove need to parse in logger once we've cleaned things up
    logger.info("Loading %s", input_emissions_file)
    input_df = pyam.IamDataFrame(input_emissions_file)

    logger.info("Converting to basic columns i.e. removing any extra columns")
    input_df = columns_to_basic(input_df)

    return input_df


def _input_checks(input_df, inputcheck, key_string, outdir):
    """
    Simple wrapper to call climate assessment workflow native emissions input
    checks, for a chosen set of checks.

    Returns IAM dataframe with only model-scenario that will be  without emissions.
    Output also includes the input emissions. The output is interpolated
    onto an annual timestep.

    For more information, see the code description under
    :func:`climate_assessment.checks.perform_input_checks`.
    """
    if inputcheck:
        LOGGER.info("Performing input data checks")
        df = perform_input_checks(
            input_df,
            output_csv_files=True,
            output_filename=key_string,
            lead_variable_check=True,
            outdir=outdir,
        )

    else:
        df = input_df.copy()

    return df


def _harmonize_and_infill(
    input_df,
    inputcheck,
    key_string,
    outdir,
    infilling_database,
    harmonize,
    prefix,
    harmonization_instance,
):
    """
    Thin wrapper function for running both harmonization and infilling.

    Returns True if there are scenarios that can be run by a climate
    emulator. Those scenarios are not returned by this function, but rather
    written to the outdir. Returns False if there are no complete scenarios
    to be run.

    For more information, see the code description under
    :func:`climate_assessment.harmonization_and_infilling.harmonization_and_infilling`.
    """
    df = _input_checks(input_df, inputcheck, key_string, outdir)

    ##################################
    # run HARMONIZATION and INFILLING (writes out results, returns True or False, includes some post-infilling checks)
    ##################################
    # TODO: split harmonization and infilling steps
    # TODO: remove 'instance' and put config somewhere more sane
    # TODO: remove prefix when we change to the new aneris interface
    assessable = harmonization_and_infilling(
        df,
        key_string,
        infilling_database,
        outdir=outdir,
        do_harmonization=harmonize,
        prefix=prefix,
        instance=harmonization_instance,
    )

    return assessable


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@input_emissions_file_arg
@outdir_arg
@input_check_option
@magicc_extra_config_option
@fair_extra_config_option
@model_option
@model_version_option
@probabilistic_file_option
@num_cfgs_option
@hist_warming_option
@hist_warming_ref_period_option
@hist_warming_eval_period_option
@test_run_option
@scenario_batch_size_option
@infilling_database_option
@save_raw_climate_output_option
@postprocess_option
@categorisation_option
@report_completeness_option
@harmonize_option
@prefix_option
@gwp_option
@harmonization_instance_option
@nonco2_warming_option
def workflow(
    input_emissions_file,
    outdir,
    inputcheck,
    magicc_extra_config,
    fair_extra_config,
    model,
    model_version,
    probabilistic_file,
    num_cfgs,
    historical_warming,
    historical_warming_reference_period,
    historical_warming_evaluation_period,
    test_run,
    scenario_batch_size,
    infilling_database,
    save_raw_climate_output,
    postprocess,
    categorisation,
    reporting_completeness_categorisation,
    harmonize,
    prefix,
    harmonization_instance,
    co2_and_non_co2_warming,
    gwp,
):
    # TODO: remove "model_version" and `num_cfgs` as mandatory
    #  options for AR6 release, as there should only be one option per emulator.
    """
    Run the full IPCC AR6 climate asessment workflow.

    Example usage: ``python scripts/run_workflow.py tests/test-data/ex2.csv output --model "FaIR" --model-version "1.6.2" --num-cfgs 2237 --probabilistic-file tests/test-data/fair-1.6.2-wg3-params-slim.json --fair-extra-config tests/test-data/fair-1.6.2-wg3-params-common.json --infilling-database src/climate_assessment/infilling/cmip6-ssps-workflow-emissions.csv``

    See further documentation at :ref:`Description of the workflow <workflow>`.
    """
    LOGGER = logging.getLogger("workflow")
    _setup_logging(LOGGER)

    run_workflow(
        input_emissions_file,
        outdir,
        model,
        model_version,
        probabilistic_file,
        num_cfgs,
        inputcheck=inputcheck,
        magicc_extra_config=magicc_extra_config,
        fair_extra_config=fair_extra_config,
        historical_warming=historical_warming,
        historical_warming_reference_period=historical_warming_reference_period,
        historical_warming_evaluation_period=historical_warming_evaluation_period,
        test_run=test_run,
        scenario_batch_size=scenario_batch_size,
        infilling_database=infilling_database,
        save_raw_climate_output=save_raw_climate_output,
        postprocess=postprocess,
        categorisation=categorisation,
        reporting_completeness_categorisation=reporting_completeness_categorisation,
        harmonize=harmonize,
        prefix=prefix,
        harmonization_instance=harmonization_instance,
        co2_and_non_co2_warming=co2_and_non_co2_warming,
        gwp=gwp,
    )


# TODO: move to a new home (probably)
def run_workflow(
    input_emissions_file,
    outdir,
    model,
    model_version,
    probabilistic_file,
    num_cfgs,
    inputcheck=True,
    magicc_extra_config=None,
    fair_extra_config=None,
    historical_warming=0.85,
    historical_warming_reference_period="1850-1900",
    historical_warming_evaluation_period="1995-2014",
    test_run=False,
    scenario_batch_size=10,
    infilling_database=os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            "infilling",
            "cmip6-ssps-workflow-emissions.csv",
        )
    ),
    save_raw_climate_output=False,
    postprocess=True,
    categorisation=True,
    reporting_completeness_categorisation=False,
    harmonize=True,
    prefix="AR6 climate diagnostics",
    harmonization_instance="ar6",
    co2_and_non_co2_warming=False,
    gwp=True,
):
    """
    Run the workflow

    Parameters
    ----------
    input_emissions_file : str
        Path to input emissions file. This must be read off disk because we use
        the filename to help name output files. PRs welcome to split off an interface
        which runs in memory.

    outdir : str
        Where to save the output

    model : str
        Climate model to run

    model_version : str
        Version of the climate model to run

    probabilistic_file : str
        File containing the probabilistic configuration of the climate model

    num_cfgs : int
        Number of climate model configurations to run

    inputcheck : bool
        Check input before running the workflow?

    magicc_extra_config : str
        File containing additional MAGICC config (only required if running MAGICC)

    fair_extra_config : str
        File containing additional FaIR config (only required if running FaIR)

    historical_warming : float
        Historical warming estimate

    historical_warming_reference_period : str
        Reference period when calculating historical warming

    historical_warming_evaluation_period : str
        Period over which historical warming is evaluated

    test_run : bool
        Are we doing a test run (if yes, we don't check that historical
        warming from the climate models comes out consistently)?

    scenario_batch_size : int
        How many scenarios to run at once (smaller number means less memory
        is needed)?

    infilling_database : str
        Path to file to use for infilling

    save_raw_climate_output : bool
        Should raw climate output be saved (warning, requires lots of disk space
        and time)?

    postprocess : bool
        Should postprocessing steps be run?

    categorisation : bool
        Should categorisation be applied to scenarios?

    reporting_completeness_categorisation : bool
        Should emissions reporting completeness be added to output?

    harmonize : bool
        Should scenarios be harmonised?

    prefix : str
        String to use as a prefix for output variables

    harmonization_instance : ["ar6", "sr15"]
        Which configuration should be used for harmonisation?

    co2_and_non_co2_warming : bool
        Calculate CO2 and non-CO2 warming too (requires 3 times as many
        climate model runs, available for MAGICC only)

    gwp : bool
        Calculate GWP equivalents too
    """
    key_string = _get_key_string_and_log_outdir(input_emissions_file, outdir, LOGGER)

    check_hist_warming_period(historical_warming_reference_period)
    check_hist_warming_period(historical_warming_evaluation_period)
    input_df = _load_emissions_convert_to_basic(input_emissions_file, LOGGER)

    assessable = _harmonize_and_infill(
        input_df,
        inputcheck,
        key_string,
        outdir,
        infilling_database,
        harmonize,
        prefix,
        harmonization_instance,
    )

    if not assessable:
        LOGGER.warning("No assessable scenarios")
        return

    # read in infilled database
    infilled_emissions = os.path.join(outdir, f"{key_string}_harmonized_infilled.csv")
    LOGGER.info("Reading in infilled scenarios from: %s", infilled_emissions)
    df_infilled = pyam.IamDataFrame(infilled_emissions)

    LOGGER.info(df_infilled.timeseries())

    ####################
    # run climate models
    ####################
    df_climate = climate_assessment(
        df_infilled,
        key_string,
        outdir,
        magicc_extra_config=magicc_extra_config,
        fair_extra_config=fair_extra_config,
        model=model,
        model_version=model_version,
        probabilistic_file=probabilistic_file,
        num_cfgs=num_cfgs,
        historical_warming=historical_warming,
        historical_warming_reference_period=historical_warming_reference_period,
        historical_warming_evaluation_period=historical_warming_evaluation_period,
        test_run=test_run,
        scenario_batch_size=scenario_batch_size,
        save_raw_output=save_raw_climate_output,
        co2_and_non_co2_warming=co2_and_non_co2_warming,
        prefix=prefix,
    )
    LOGGER.info(df_climate.timeseries())

    LOGGER.info("Concatenating infilled df, climate df and input df")
    results = pyam.concat([df_infilled, df_climate])
    output = pyam.concat(
        [input_df.filter(variable=results.variable, keep=False), results]
    )

    LOGGER.info("write out raw output")
    output.to_excel(os.path.join(outdir, str(key_string + "_" + "rawoutput.xlsx")))

    if postprocess:
        output_postprocess = do_postprocess(
            output,
            outdir=outdir,
            key_string=key_string,
            categorisation=categorisation,
            reporting_completeness_categorisation=reporting_completeness_categorisation,
            prefix=prefix,
            gwp=gwp,
            model_version=model_version,
            model=model,
        )

        # Sanity checks
        sanity_check_bounds_kyoto_emissions(
            output_postprocess,
            out_kyoto_infilled=f"{prefix}|Infilled|Emissions|Kyoto Gases",
        )
        sanity_check_comparison_kyoto_gases(
            output_postprocess,
            out_kyoto_harmonized=f"{prefix}|Harmonized|Emissions|Kyoto Gases",
            out_kyoto_infilled=f"{prefix}|Infilled|Emissions|Kyoto Gases",
        )


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@input_emissions_file_arg
@outdir_arg
@input_check_option
@infilling_database_option
@prefix_option
@harmonization_instance_option
def harmonize_and_infill(
    input_emissions_file,
    outdir,
    inputcheck,
    infilling_database,
    prefix,
    # gwp,
    harmonization_instance,
):
    """
    Harmonise and infill data in ``input_emissions_file``, saving output in ``outdir``

    ``input_emissions_file`` should be a path to a file of emissions

    ``outdir`` should be a path a directory which already exists
    """
    LOGGER = logging.getLogger("harmonize_and_infill")
    _setup_logging(LOGGER)

    key_string = _get_key_string_and_log_outdir(input_emissions_file, outdir, LOGGER)

    input_df = _load_emissions_convert_to_basic(input_emissions_file, LOGGER)

    _harmonize_and_infill(
        input_df=input_df,
        inputcheck=inputcheck,
        key_string=key_string,
        outdir=outdir,
        infilling_database=infilling_database,
        prefix=prefix,
        # gwp=gwp,  # TODO: implement downstream
        harmonization_instance=harmonization_instance,
        harmonize=True,
    )


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@input_emissions_file_arg
@outdir_arg
@input_check_option
@prefix_option
@gwp_option
@harmonization_instance_option
def harmonize(
    input_emissions_file,
    outdir,
    inputcheck,
    prefix,
    gwp,
    harmonization_instance,
):
    """
    Harmonise data in ``input_emissions_file``, saving output in ``outdir``

    ``input_emissions_file`` should be a path to a file of emissions

    ``outdir`` should be a path a directory which already exists
    """
    LOGGER = logging.getLogger("harmonize")
    _setup_logging(LOGGER)

    key_string = _get_key_string_and_log_outdir(input_emissions_file, outdir, LOGGER)

    input_df = _load_emissions_convert_to_basic(input_emissions_file, LOGGER)

    df = _input_checks(input_df, inputcheck, key_string, outdir)

    df_harmonized = run_harmonization(
        df, instance=harmonization_instance, prefix=prefix
    )

    if gwp:
        df_harmonized = add_gwp100_kyoto_wrapper(
            df_harmonized,
            gwps=["AR5GWP100", "AR6GWP100"],
            prefixes=[f"{prefix}|Harmonized|"],
        )

    df_harmonized.to_csv(os.path.join(outdir, f"{key_string}_harmonized.csv"))


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@input_emissions_file_arg
@outdir_arg
@input_check_option
@prefix_option
@gwp_def_false_option
@harmonization_instance_option
def create_infiller_database(
    input_emissions_file,
    outdir,
    inputcheck,
    prefix,
    gwp,
    harmonization_instance,
):
    """
    Creates infiller database by harmonizing data in ``input_emissions_file``
    which is followed by some common-sense vetting and then saved in ``outdir``

    ``input_emissions_file`` should be a path to a file of emissions

    ``outdir`` should be a path a directory which already exists
    """
    LOGGER = logging.getLogger("create_infiller_database")
    _setup_logging(LOGGER)

    key_string = _get_key_string_and_log_outdir(input_emissions_file, outdir, LOGGER)

    input_df = _load_emissions_convert_to_basic(input_emissions_file, LOGGER)

    df = _input_checks(input_df, inputcheck, key_string, outdir)

    df_harmonized = run_harmonization(
        df, instance=harmonization_instance, prefix=prefix
    )

    df_infiller_database = infiller_vetting(df_harmonized)

    if gwp:
        df_infiller_database = add_gwp100_kyoto_wrapper(
            df_infiller_database,
            gwps=["AR5GWP100", "AR6GWP100"],
            prefixes=[f"{prefix}|Harmonized|"],
        )

    # Calculate total emissions consistent with afolu and energy for the
    # infilling database to avoid losing scenarios which have total, energy
    # and afolu reported and total dropped in the checks stage. We could also
    # do this at end of harmonization too, depending on whether we want
    # the infilling to preference using total CO2 or only fossil CO2
    LOGGER.info("Calculating CO2 totals where missing")
    _, no_co2_totals = split_df(
        df_infiller_database.filter(variable="*Harmonized*CO2*"),
        variable=f"{prefix}|Harmonized|Emissions|CO2",
    )
    co2_totals = _add_variables(
        no_co2_totals,
        f"{prefix}|Harmonized|Emissions|CO2|Energy and Industrial Processes",
        f"{prefix}|Harmonized|Emissions|CO2|AFOLU",
        f"{prefix}|Harmonized|Emissions|CO2",
        raise_if_mismatch=False,
    )
    df_infiller_database = df_infiller_database.append(co2_totals, inplace=False)

    LOGGER.info("Saving output")
    df_infiller_database.to_csv(
        os.path.join(outdir, f"{key_string}_infillerdatabase.csv")
    )


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@input_emissions_file_arg
@outdir_arg
@infilling_database_option
@prefix_option
@gwp_option
def infill(
    input_emissions_file,
    outdir,
    infilling_database,
    prefix,
    gwp,
):
    """
    Infill harmonized data in ``input_emissions_file``, saving output in ``outdir``

    ``input_emissions_file`` should be a path to a file of harmonized emissions

    ``outdir`` should be a path a directory which already exists
    """
    LOGGER = logging.getLogger("infill")
    _setup_logging(LOGGER)

    LOGGER.info("Loading harmonized emissions from %s", input_emissions_file)
    LOGGER.info("Output will be saved in %s", outdir)
    harmonized = pyam.IamDataFrame(input_emissions_file)
    harmonized_start_year = min(harmonized.year)

    LOGGER.info("Calling infilling")
    infilled, _, _ = run_infilling(
        harmonized,
        prefix=prefix,
        database_filepath=infilling_database,
        start_year=harmonized_start_year,
    )

    LOGGER.info("Post-processing for the climate model step")
    infilled = postprocess_infilled_for_climate(
        infilled, prefix=prefix, start_year=harmonized_start_year
    )

    if gwp:
        infilled = add_gwp100_kyoto_wrapper(
            infilled,
            gwps=["AR5GWP100", "AR6GWP100"],
            prefixes=[f"{prefix}|Infilled|"],
        )

    outfile = os.path.join(outdir, "infilled.csv")
    LOGGER.info("Saving output to %s", outfile)
    infilled.to_csv(outfile)


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@harmonizedinfilledemissions_arg
@outdir_arg
@num_cfgs_option
@hist_warming_option
@hist_warming_ref_period_option
@hist_warming_eval_period_option
@test_run_option
@model_option
@model_version_option
@magicc_extra_config_option
@fair_extra_config_option
@probabilistic_file_option
@scenario_batch_size_option
@prefix_option
@gwp_def_false_option
@nonco2_warming_option
@save_raw_climate_output_option
@save_csv_combined_output_option
def clim_cli(
    harmonizedinfilledemissions,
    outdir,
    num_cfgs,
    historical_warming,
    historical_warming_reference_period,
    historical_warming_evaluation_period,
    test_run,
    model,
    model_version,
    magicc_extra_config,
    fair_extra_config,
    probabilistic_file,
    scenario_batch_size,
    prefix,
    gwp,
    co2_and_non_co2_warming,
    save_raw_climate_output,
    save_csv_combined_output,
):
    """
    Run the climate emulator step of the IPCC AR6 climate asessment workflow.

    Example usage: ``python scripts/run_clim.py tests/test-data/workflow-fair/ex2_harmonized_infilled.csv output --model "fair" --model-version "1.6.2" --num-cfgs 2237 --probabilistic-file tests/test-data/fair-1.6.2-wg3-params-slim.json --fair-extra-config tests/test-data/fair-1.6.2-wg3-params-common.json --scenario-batch-size 4``

    For more information, see the code description under
    :func:`climate_assessment.climate.climate_assessment`.
    """
    LOGGER = logging.getLogger("clim_cli")
    _setup_logging(LOGGER)

    # silence pyam loggers
    logging.getLogger("pyam.core").setLevel(logging.CRITICAL)
    logging.getLogger("pyam.utils").setLevel(logging.CRITICAL)

    LOGGER.info("Outputs will be saved in: %s", outdir)

    key_string = os.path.splitext(os.path.basename(harmonizedinfilledemissions))[0]
    LOGGER.info("Outputs will be saved with the ID: %s", key_string)

    check_hist_warming_period(historical_warming_reference_period)
    check_hist_warming_period(historical_warming_evaluation_period)

    LOGGER.info(
        "Loading harmonized, infilled scenarios: %s", harmonizedinfilledemissions
    )
    df_infilled = pyam.IamDataFrame(harmonizedinfilledemissions)
    df_climate = climate_assessment(
        df_infilled,
        key_string,
        outdir,
        magicc_extra_config=magicc_extra_config,
        fair_extra_config=fair_extra_config,
        model=model,
        model_version=model_version,
        probabilistic_file=probabilistic_file,
        num_cfgs=num_cfgs,
        historical_warming=historical_warming,
        historical_warming_reference_period=historical_warming_reference_period,
        historical_warming_evaluation_period=historical_warming_evaluation_period,
        test_run=test_run,
        scenario_batch_size=scenario_batch_size,
        save_raw_output=save_raw_climate_output,
        co2_and_non_co2_warming=co2_and_non_co2_warming,
        prefix=prefix,
    )
    if df_climate is None:
        LOGGER.error("Climate assessment failed, exiting")

        return

    LOGGER.info("Concatenating infilled df and climate df")
    results = pyam.concat([df_infilled, df_climate])

    if gwp:
        LOGGER.info("Adding extra Kyoto Gases variables in GWP100 for each scenario")
        results = add_gwp100_kyoto_wrapper(
            results,
            gwps=["AR5GWP100", "AR6GWP100"],
            prefixes=[
                "",
                "AR6 climate diagnostics|Harmonized|",
                "AR6 climate diagnostics|Infilled|",
            ],
        )

    LOGGER.info("write out raw output")
    results.to_excel(os.path.join(outdir, str(key_string + "_" + "rawoutput.xlsx")))

    if save_csv_combined_output:
        LOGGER.info("write out raw output in csv")
        results.to_csv(os.path.join(outdir, str(key_string + "_" + "rawoutput.csv")))

    LOGGER.info("COMPLETE")


def _postprocess_worker(fname, outdir, **kwargs):
    """
    Helper function which takes a file that ends with "_rawoutput.xlsx" and the output
    location, and then calls the function `do_postprocess(output, outdir, key_string, prefix)`.
    """
    if not fname.endswith("_rawoutput.xlsx"):
        raise AssertionError(fname)

    output = pyam.IamDataFrame(fname)

    key_string = os.path.basename(fname).replace("_rawoutput.xlsx", "")

    return do_postprocess(output, outdir=outdir, key_string=key_string, **kwargs)


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@output_files_arg
@outdir_arg
@n_workers_option
@output_file_option
@prefix_all_out_variables_option
@kyoto_ghgs_option
@model_option
@model_version_option
@categorisation_option
@report_completeness_option
def postprocess(
    rawoutput_files,
    outdir,
    output_file,
    prefix,
    kyoto_ghgs,
    n_workers,
    model,
    model_version,
    categorisation,
    reporting_completeness_categorisation,
):
    """
    Merge and postprocess a collection of rawoutput into a single set of output files

    This script should be run after run_workflow.py if --no-postprocess was supplied. This
    is particularly useful for a clusted run where a run_workflow is run multiple times with
    batches of scenario data.

    Each file is processed individually, but the metadata from all the runs is merged into a single
    output file.
    """
    LOGGER = logging.getLogger("clim_cli")
    _setup_logging(LOGGER)

    LOGGER.info(f"Merging {len(rawoutput_files)} files: {rawoutput_files}")
    LOGGER.info(f"Saving output as {os.path.join(outdir, output_file)}")

    with ProcessPoolExecutor(max_workers=n_workers) as executor:
        futures = [
            executor.submit(
                _postprocess_worker,
                fname,
                outdir,
                model=model,
                model_version=model_version,
                categorisation=categorisation,
                reporting_completeness_categorisation=reporting_completeness_categorisation,
                gwp=kyoto_ghgs,
                prefix=prefix,
            )
            for fname in rawoutput_files
        ]

    results = []
    for r in tqdm(as_completed(futures), total=len(futures)):
        if r is not None:
            results.append(r.result().meta)

    out_fname = os.path.join(outdir, output_file)
    LOGGER.info(f"Saving merged meta to {out_fname}")
    pd.concat(results).to_csv(out_fname)


batch_size_option = click.option(
    "--batch-size",
    help="Maximum number of scenarios per batch",
    required=True,
    default=20,
    type=int,
    show_default=True,
)


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@input_emissions_file_arg
@outdir_arg
@batch_size_option
def _split_scenarios_into_batches(input_emissions_file, outdir, batch_size):
    """
    Split scenario data into multiple batch files with at maximum the number of
    scenarios specified in batch_size per file.
    """
    split_scenarios_into_batches(
        iamc_file=input_emissions_file, outdir=outdir, batch_size=batch_size
    )


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@input_emissions_file_arg
@outdir_arg
def _extract_ips(input_emissions_file, outdir):
    """
    Extract IP scenarios from a scenario file.
    """
    extract_ips(ar6_file=input_emissions_file, outdir=outdir)



import json
import logging
import os.path
from functools import lru_cache

import numpy as np
import numpy.testing as npt
import pandas as pd
import scmdata
import scmdata.database
import scmdata.processing
from pint.errors import DimensionalityError

from .ciceroscm import ciceroscm_post_process
from .fair import fair_post_process
from .magicc7 import calculate_co2_and_nonco2_warming_magicc, magicc7_post_process

LOGGER = logging.getLogger(__name__)
_CLIMATE_VARIABLE_DEFINITION_CSV = os.path.join(
    os.path.dirname(__file__), "variable_definitions.csv"
)


@lru_cache
def _get_climate_variable_definitions(fname):
    return pd.read_csv(fname)


def convert_openscm_runner_variables_to_ar6_wg3_variables(in_var):
    mapping = {
        "Surface Air Temperature Change": "Raw Surface Temperature (GSAT)",
        "Surface Air Ocean Blended Temperature Change": "Raw Surface Temperature (GMST)",
        "Heat Uptake|Ocean": "Ocean Heat Uptake",
        "Effective Radiative Forcing|Aerosols|Direct Effect|SOx": "Effective Radiative Forcing|Aerosols|Direct Effect|Sulfur",
    }

    try:
        return mapping[in_var]
    except KeyError:
        return in_var


def check_hist_warming_period(period):
    """
    Check period for historical warming calculations

    Parameters
    ----------
    period : str
        Input period, must be formatted as "YYYY-YYYY" where the first year is
        less than or equal to the last year. The last year is included in the
        period.

    Returns
    -------
    range
        All years in the period

    Raises
    ------
    ValueError
        ``period`` is formatted incorrectly
    """

    def _raise_error():
        raise ValueError(
            f"`period` must be a string of the form 'YYYY-YYYY' (with the first year "
            f"being less than or equal to the second), we received {period}"
        )

    if len(period) != 9:
        _raise_error()

    try:
        split = period.split("-")
    except TypeError:
        _raise_error()

    if len(split) != 2:
        _raise_error()

    try:
        start_year = int(split[0])
        end_year = int(split[1])
    except TypeError:
        _raise_error()

    if start_year > end_year:
        _raise_error()

    return range(start_year, end_year + 1)


def calculate_exceedance_probability_timeseries(
    res,
    exceedance_probability_calculation_var,
    test_run=False,
    historical_warming=0.85,
    historical_warming_reference_period="1850-1900",
    historical_warming_evaluation_period="1995-2014",
):
    """
    Calculate the timeseries with which we should determine exceedance probabilities.

    Please note that calculating the statistical properties like the exceedance probability is done on the count of the raw, full ensemble data.
    It  cannot be derived from from the time series of the final statistical indicator afterwards.
    """
    hist_temp_ref_period = check_hist_warming_period(
        historical_warming_reference_period
    )
    hist_temp_evaluation_period = check_hist_warming_period(
        historical_warming_evaluation_period
    )
    historical_warming_unit = "K"

    LOGGER.info(
        "Adjusting median of %s-%s warming (rel. to %s-%s) to %s%s",
        hist_temp_evaluation_period[0],
        hist_temp_evaluation_period[-1],
        hist_temp_ref_period[0],
        hist_temp_ref_period[-1],
        historical_warming,
        historical_warming_unit,
    )

    exceedance_probability_timeseries_raw = res.filter(
        variable=f"Raw {exceedance_probability_calculation_var}",
        unit=historical_warming_unit,
    )
    exceedance_probability_timeseries_rel_ref_period = (
        exceedance_probability_timeseries_raw.relative_to_ref_period_mean(
            year=hist_temp_ref_period
        ).drop_meta(["reference_period_start_year", "reference_period_end_year"])
    )
    exceedance_probability_timeseries = (
        exceedance_probability_timeseries_rel_ref_period.adjust_median_to_target(
            historical_warming,
            hist_temp_evaluation_period,
            process_over=("run_id",),
        )
    )

    # output checks
    hist_temp_grouper_cols = exceedance_probability_timeseries.get_meta_columns_except(
        "run_id"
    )

    def _get_median_hist_warming(inp):
        return (
            inp.filter(year=hist_temp_evaluation_period)
            .timeseries()
            .mean(axis="columns")
            .groupby(hist_temp_grouper_cols)
            .median()
        )

    inp_vals = _get_median_hist_warming(exceedance_probability_timeseries_raw)
    check_vals = _get_median_hist_warming(exceedance_probability_timeseries)
    shifts = check_vals - inp_vals
    shifts_after_rebase = check_vals - _get_median_hist_warming(
        exceedance_probability_timeseries_rel_ref_period
    )

    if not np.isclose(shifts, shifts[0], atol=5 * 1e-3).all():
        LOGGER.exception(
            "Careful of scenarios which break match with history! `shifts`: %s",
            shifts,
        )
    else:
        LOGGER.info("`shifts`: %s", shifts)
        LOGGER.info("`shifts_after_rebase`: %s", shifts_after_rebase)

    if not test_run:
        try:
            npt.assert_allclose(
                check_vals,
                historical_warming,
                rtol=1e-2,  # to within 1%
                err_msg=f"{check_vals}",
            )
        except AssertionError:
            LOGGER.exception("Careful of scenarios which break match with history!")

    exceedance_probability_timeseries["variable"] = (
        exceedance_probability_calculation_var
    )
    return exceedance_probability_timeseries


def calculate_co2_and_nonco2_warming_and_remove_extras(res):
    """
    Calculate non-CO2 warming. Currently only implemented for MAGICC, which can
    be run twice to allow for this calculation, using (the CLI option)
    `co2_and_non_co2_warming`.
    """
    out = []
    for cmrun in res.groupby("climate_model"):
        climate_model = cmrun.get_unique_meta("climate_model", no_duplicates=True)
        if climate_model.startswith("MAGICC"):
            all_forcers_run = cmrun.filter(rf_total_runmodus="ALL").drop_meta(
                "rf_total_runmodus"
            )
            out.append(all_forcers_run)
            out.append(calculate_co2_and_nonco2_warming_magicc(cmrun))
        else:
            raise NotImplementedError(climate_model)

    return scmdata.run_append(out)


def post_process(
    res,
    outdir,
    test_run=False,
    save_raw_output=False,
    co2_and_non_co2_warming=False,
    # for exceedance probability calculations
    temp_thresholds=(1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0),
    peak_percentiles=(5, 10, 17, 25, 33, 50, 66, 67, 75, 83, 90, 95),
    percentiles=(
        5,
        10,
        1 / 6 * 100,
        17,
        25,
        33,
        50,
        66,
        67,
        75,
        83,
        5 / 6 * 100,
        90,
        95,
    ),
    historical_warming=0.85,
    historical_warming_reference_period="1850-1900",
    historical_warming_evaluation_period="1995-2014",
):
    LOGGER.info("Beginning climate post-processing")
    LOGGER.info("Removing unknown units and keeping only World data")
    res = res.filter(unit="unknown", keep=False).filter(region="World")

    LOGGER.info(
        "Renaming variables from OpenSCM-Runner conventions to AR6 WG3 conventions"
    )
    res["variable"] = res["variable"].apply(
        convert_openscm_runner_variables_to_ar6_wg3_variables
    )

    LOGGER.info("Performing climate model specific fixes")
    all_res = []
    for res_cm in res.groupby("climate_model"):
        climate_model = res_cm.get_unique_meta("climate_model", no_duplicates=True)
        LOGGER.info("Processing %s data", climate_model)

        if climate_model.startswith("MAGICCv7"):
            res_cm = magicc7_post_process(res_cm)

        if climate_model.startswith("FaIRv1.6"):
            res_cm = fair_post_process(res_cm)

        if climate_model.startswith("CICERO-SCM"):
            res_cm = ciceroscm_post_process(res_cm)

        all_res.append(res_cm)

    LOGGER.info("Recombining post-processed data")
    res = scmdata.run_append(all_res)

    def _rename_vars(v):
        mapping = {
            "Effective Radiative Forcing|Greenhouse Gases": "Effective Radiative Forcing|Basket|Greenhouse Gases",
            "Effective Radiative Forcing|Anthropogenic": "Effective Radiative Forcing|Basket|Anthropogenic",
        }

        try:
            out = mapping[v]
            LOGGER.debug("Renaming %s to %s", v, out)
            return out
        except KeyError:
            LOGGER.debug("Not renaming %s", v)
            return v

    LOGGER.info("Performing further variable renaming")
    res["variable"] = res["variable"].apply(_rename_vars)

    if save_raw_output:
        LOGGER.info("Saving raw output (with renamed variables) to disk")
        if "parameters" in res.metadata:
            res.metadata["parameters"] = json.dumps(res.metadata["parameters"])

        database = scmdata.database.ScmDatabase(
            os.path.join(outdir, "raw_climate_output"),
            levels=["climate_model", "model", "scenario"],
        )
        for c in [
            "climate_model",
            "model",
            "scenario",
            "variable",
            "region",
            "unit",
            "rf_total_runmodus",
        ]:
            if c in res.meta:
                res[c] = res[c].apply(str)

        # TODO: add test for save raw output with non-CO2 on
        database.save(res)

    if co2_and_non_co2_warming:
        LOGGER.info("Calculating non-CO2 warming")
        res = calculate_co2_and_nonco2_warming_and_remove_extras(res)

    LOGGER.info("Calculating exceedance probability timeseries")
    exceedance_probability_calculation_var = "Surface Temperature (GSAT)"
    exceedance_probability_timeseries = calculate_exceedance_probability_timeseries(
        res,
        exceedance_probability_calculation_var,
        test_run=test_run,
        historical_warming=historical_warming,
        historical_warming_reference_period=historical_warming_reference_period,
        historical_warming_evaluation_period=historical_warming_evaluation_period,
    )
    res = res.append(exceedance_probability_timeseries)

    year_filter = range(1995, 2101)
    LOGGER.info("Keeping only data from %s", year_filter)
    res = res.filter(year=year_filter)

    LOGGER.info("Calculating Non-CO2 GHG ERF")
    helper = res.filter(variable="Effective Radiative Forcing*")
    res = [res]

    erf_nonco2_ghg = helper.filter(
        variable="Effective Radiative Forcing|Basket|Greenhouse Gases"
    ).subtract(
        helper.filter(variable="Effective Radiative Forcing|CO2"),
        op_cols={
            "variable": "Effective Radiative Forcing|Basket|Non-CO2 Greenhouse Gases"
        },
    )

    if (
        erf_nonco2_ghg.get_unique_meta("unit", no_duplicates=True)
        != "watt / meter ** 2"
    ):
        raise AssertionError("Unexpected forcing unit")

    erf_nonco2_ghg["unit"] = "W/m^2"
    res.append(erf_nonco2_ghg)

    LOGGER.info("Calculating Non-CO2 Anthropogenic ERF")
    erf_nonco2_anthropogenic = helper.filter(
        variable="Effective Radiative Forcing|Basket|Anthropogenic"
    ).subtract(
        helper.filter(variable="Effective Radiative Forcing|CO2"),
        op_cols={
            "variable": "Effective Radiative Forcing|Basket|Non-CO2 Anthropogenic"
        },
    )
    if (
        erf_nonco2_anthropogenic.get_unique_meta("unit", no_duplicates=True)
        != "watt / meter ** 2"
    ):
        raise AssertionError("Unexpected forcing unit")

    erf_nonco2_anthropogenic["unit"] = "W/m^2"
    res.append(erf_nonco2_anthropogenic)

    LOGGER.info("Joining derived variables and data back together")
    res = scmdata.run_append(res)

    # check all variable names
    LOGGER.info("Converting all variable names and units to standard definitions")

    def _convert_to_standard_name_and_unit(vdf):
        climate_variable_definitions = _get_climate_variable_definitions(
            _CLIMATE_VARIABLE_DEFINITION_CSV
        )
        variable = vdf.get_unique_meta("variable", True)

        try:
            standard_unit = climate_variable_definitions.set_index("Variable").loc[
                variable
            ]["Unit"]
        except KeyError as exc:
            raise ValueError(
                f"{variable} not in {_CLIMATE_VARIABLE_DEFINITION_CSV}"
            ) from exc
        try:
            return vdf.convert_unit(standard_unit)
        except DimensionalityError as exc:
            raise ValueError(
                "Cannot convert {} units of {} to {}".format(
                    variable, vdf.get_unique_meta("unit", True), standard_unit
                )
            ) from exc

    res = res.groupby("variable").map(_convert_to_standard_name_and_unit)

    LOGGER.info("Calculating percentiles")
    res_percentiles = res.quantiles_over(
        "run_id", np.array(percentiles) / 100
    ).reset_index()
    res_percentiles["percentile"] = res_percentiles["quantile"] * 100
    res_percentiles = res_percentiles.drop("quantile", axis="columns")

    LOGGER.info("Mangling variable name with climate model and percentile")
    res_percentiles["variable"] = (
        res_percentiles["variable"].astype(str)
        + "|"
        + res_percentiles["climate_model"].astype(str)
        + "|"
        + res_percentiles["percentile"].astype(float).round(1).astype(str)
        + "th Percentile"
    )
    res_percentiles = scmdata.ScmRun(
        res_percentiles.drop(["climate_model", "percentile"], axis="columns")
    )

    LOGGER.info(
        "Calculating exceedance probabilities and exceedance probability timeseries"
    )
    res_exceedance_probs_var = res.filter(
        variable=exceedance_probability_calculation_var
    )
    exceedance_probs_by_temp_threshold = []
    exceedance_probs_tss = []
    for threshold in temp_thresholds:
        LOGGER.info("Calculating %s exceedance probabilities", threshold)
        for store, func, output_name in (
            (
                exceedance_probs_by_temp_threshold,
                scmdata.processing.calculate_exceedance_probabilities,
                f"Exceedance Probability {threshold}C",
            ),
            (
                exceedance_probs_tss,
                scmdata.processing.calculate_exceedance_probabilities_over_time,
                f"Exceedance Probability {threshold}C",
            ),
        ):
            store.append(
                func(
                    res_exceedance_probs_var,
                    threshold,
                    process_over_cols=("run_id",),
                    output_name=output_name,
                )
            )

    exceedance_probs_tss = pd.concat(exceedance_probs_tss).reset_index()
    exceedance_probs_tss["variable"] = (
        exceedance_probs_tss["variable"].astype(str)
        + "|"
        + exceedance_probs_tss["climate_model"].astype(str)
    )
    exceedance_probs_tss = exceedance_probs_tss.drop("climate_model", axis="columns")
    res_percentiles = res_percentiles.append(exceedance_probs_tss)

    reporting_groups = ["climate_model", "model", "scenario"]
    exceedance_probs_by_temp_threshold = pd.concat(
        exceedance_probs_by_temp_threshold, axis=1
    )
    exceedance_probs_by_temp_threshold = exceedance_probs_by_temp_threshold.reset_index(
        list(
            set(exceedance_probs_by_temp_threshold.index.names) - set(reporting_groups)
        ),
        drop=True,
    )

    LOGGER.info("Calculating peak warming and peak warming year")
    peaks = scmdata.processing.calculate_peak(res_exceedance_probs_var)
    peak_years = scmdata.processing.calculate_peak_time(res_exceedance_probs_var)

    def _get_quantiles(idf):
        return (
            idf.groupby(reporting_groups)
            .quantile(np.array(peak_percentiles) / 100)
            .unstack()
        )

    peaks_quantiles = _get_quantiles(peaks)
    peak_years_quantiles = _get_quantiles(peak_years).astype(int)

    def rename_quantiles(quantile):
        percentile = int(quantile * 100)
        if np.isclose(percentile, 50):
            plabel = "median"
        else:
            plabel = f"p{percentile}"

        return plabel

    peaks_quantiles.columns = (
        peaks_quantiles.columns.map(rename_quantiles).astype(str) + " peak warming"
    )
    peak_years_quantiles.columns = (
        peak_years_quantiles.columns.map(rename_quantiles).astype(str)
        + " year of peak warming"
    )

    LOGGER.info("Creating meta table")
    meta_table = pd.concat(
        [
            exceedance_probs_by_temp_threshold,
            peaks_quantiles,
            peak_years_quantiles,
        ],
        axis=1,
    )

    def mangle_meta_table_climate_model(idf):
        out = idf.copy()
        climate_model = idf.name
        out.columns = out.columns + f" ({climate_model})"

        return out

    meta_table = (
        meta_table.groupby("climate_model", group_keys=False)
        .apply(mangle_meta_table_climate_model)
        .reset_index("climate_model", drop=True)
    )

    LOGGER.info("Exiting post-processing")
    return res, res_percentiles, meta_table



import json
import logging

import numpy as np
from openscm_runner.adapters import FAIR

LOGGER = logging.getLogger(__name__)
DEFAULT_FAIR_VERSION = "1.6.2"


def get_fair_configurations(
    fair_version, fair_probabilistic_file, fair_extra_config, num_cfgs
):
    """
    Get configuration for FaIR
    """
    if FAIR.get_version() != fair_version:
        raise AssertionError(FAIR.get_version())

    with open(fair_probabilistic_file) as fh:
        cfgs_raw = json.load(fh)

    with open(fair_extra_config) as fh:
        cfgs_common = json.load(fh)

    e_pi = [0] * 40
    for idx in range(5, 12):
        e_pi[idx] = cfgs_common["E_pi"][idx - 5]

    fair_cfgs = []
    for i, c in enumerate(cfgs_raw[:num_cfgs]):
        scale = [1] * 45
        c_pi = [0] * 31
        c_pi[1] = cfgs_common["C_pi"][0]
        c_pi[2] = cfgs_common["C_pi"][1]
        c_pi[3] = cfgs_common["C_pi"][2]
        c_pi[20] = cfgs_common["C_pi"][3]
        c_pi[25] = cfgs_common["C_pi"][4]
        c_pi[29] = cfgs_common["C_pi"][5]
        c_pi[30] = cfgs_common["C_pi"][6]
        scale[1] = c["scale"][0]
        scale[2] = c["scale"][1]
        for idx in range(3, 31):
            scale[idx] = c["scale"][2]
        scale[15] = scale[15] * cfgs_common["cfc11_adj"]
        scale[16] = scale[16] * cfgs_common["cfc12_adj"]
        scale[33] = c["scale"][3]
        scale[34] = c["scale"][4]
        scale[41] = c["scale"][5]
        scale[42] = c["scale"][6]
        scale[43] = c["scale"][7]
        c_pi[0] = c["C_pi_CO2"]
        f_solar = np.zeros(361)
        f_solar[:270] = (
            np.linspace(0, c["trend_solar"], 270)
            + np.array(cfgs_common["default_solar"])[:270] * c["scale"][8]
        )
        f_solar[270:351] = (
            c["trend_solar"]
            + np.array(cfgs_common["default_solar"])[270:351] * c["scale"][8]
        )
        f_solar[351:361] = cfgs_common["default_solar"][351:]
        this_cfg = {
            "run_id": i,
            "F2x": c["F2x"],
            "r0": c["r0"],
            "rt": c["rt"],
            "rc": c["rc"],
            "lambda_global": c["lambda_global"],
            "ocean_heat_capacity": c["ocean_heat_capacity"],
            "ocean_heat_exchange": c["ocean_heat_exchange"],
            "deep_ocean_efficacy": c["deep_ocean_efficacy"],
            "b_aero": [
                c["b_aero"][0],
                0.0,
                0.0,
                0.0,
                c["b_aero"][1],
                c["b_aero"][2],
                c["b_aero"][3],
            ],
            "ghan_params": c["ghan_params"],
            "scale": scale,
            "F_solar": f_solar.tolist(),
            "F_volcanic": cfgs_common["default_volcanic"],
            "C_pi": c_pi,
            "b_tro3": c["b_tro3"],
            "ozone_feedback": c["ozone_feedback"],
            "E_pi": e_pi,
            "ghg_forcing": cfgs_common["ghg_forcing"],
            "aCO2land": cfgs_common["aCO2land"],
            "stwv_from_ch4": cfgs_common["stwv_from_ch4"],
            "F_ref_BC": cfgs_common["F_ref_BC"],
            "E_ref_BC": cfgs_common["E_ref_BC"],
            "tropO3_forcing": cfgs_common["tropO3_forcing"],
            "natural": cfgs_common["natural"],
        }
        fair_cfgs.append(this_cfg)

    return fair_cfgs


def fair_post_process(climate_output):
    # convert units to W/m^2
    climate_output = climate_output.convert_unit(
        "W/m^2", variable="Effective Radiative Forcing*"
    )

    return climate_output
